#!/usr/bin/env python
#
# Glances - An eye on your system
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#
# Refactor by @ariel-anieli in 2024

"""Glances unitary tests suite."""

import json
import multiprocessing
import time
import unittest
from datetime import datetime
from unittest.mock import patch

# Ugly hack waiting for Python 3.10 deprecation
try:
    from datetime import UTC
except ImportError:
    from datetime import timezone

    UTC = timezone.utc

from glances import __version__
from glances.events_list import GlancesEventsList
from glances.filter import GlancesFilter, GlancesFilterList
from glances.globals import (
    BSD,
    LINUX,
    MACOS,
    SUNOS,
    WINDOWS,
    WSL,
    auto_unit,
    pretty_date,
    split_esc,
    string_value_to_float,
    subsample,
)
from glances.main import GlancesMain
from glances.outputs.glances_bars import Bar
from glances.plugins.fs.zfs import zfs_enable, zfs_stats
from glances.plugins.mpp import MppPlugin
from glances.plugins.npu import NpuPlugin
from glances.plugins.plugin.dag import get_plugin_dependencies
from glances.plugins.plugin.model import GlancesPluginModel
from glances.stats import GlancesStats
from glances.thresholds import (
    GlancesThresholdCareful,
    GlancesThresholdCritical,
    GlancesThresholdOk,
    GlancesThresholds,
    GlancesThresholdWarning,
)

# Multiprocessing start method (on POSIX system)
if LINUX or BSD or SUNOS or MACOS:
    ctx_mp_fork = multiprocessing.get_context('fork')
else:
    ctx_mp_fork = multiprocessing.get_context()

# Global variables
# =================

# Init Glances core
testargs = ["glances", "-C", "./conf/glances.conf"]
with patch('sys.argv', testargs):
    core = GlancesMain()
test_config = core.get_config()
test_args = core.get_args()

# Init Glances stats
stats = GlancesStats(config=test_config, args=test_args)

# Unitest class
# ==============
print(f'Unitary tests for Glances {__version__}')


class TestGlances(unittest.TestCase):
    """Test Glances class."""

    def setUp(self):
        """The function is called *every time* before test_*."""
        print('\n' + '=' * 78)

    def _common_plugin_tests(self, plugin):
        """Common method to test a Glances plugin
        This method is called multiple time by test 100 to 1xx"""

        assert stats.args.conf_file == './conf/glances.conf', 'Configuration file not correctly set in stats'
        # But not take into account

        # Reset all the stats, history and views
        plugin_instance = stats.get_plugin(plugin)
        plugin_instance.reset()  # reset stats
        plugin_instance.reset_views()  # reset views
        plugin_instance.reset_stats_history()  # reset history

        # Check before update
        self.assertEqual(plugin_instance.get_raw(), plugin_instance.stats_init_value)
        self.assertEqual(plugin_instance.is_enabled(), True)
        self.assertEqual(plugin_instance.is_disabled(), False)
        self.assertEqual(plugin_instance.get_views(), {})
        self.assertIsInstance(plugin_instance.get_raw(), (dict, list))
        if plugin_instance.history_enable() and isinstance(plugin_instance.get_raw(), dict):
            self.assertEqual(plugin_instance.get_key(), None)
            self.assertTrue(
                all(
                    f in [h['name'] for h in plugin_instance.items_history_list]
                    for f in plugin_instance.get_raw_history()
                )
            )
        elif plugin_instance.history_enable() and isinstance(plugin_instance.get_raw(), list):
            self.assertNotEqual(plugin_instance.get_key(), None)

        # Update stats (add first element)
        plugin_instance.update()
        plugin_instance.update_stats_history()
        plugin_instance.update_views()

        # Check stats
        self.assertIsInstance(plugin_instance.get_raw(), (dict, list))
        if isinstance(plugin_instance.get_raw(), dict) and plugin_instance.get_raw() != {}:
            res = False
            for f in plugin_instance.fields_description:
                if f not in plugin_instance.get_raw():
                    print(f"WARNING: {f} field not found in {plugin} plugin stats")
                else:
                    res = True
            self.assertTrue(res)
        elif isinstance(plugin_instance.get_raw(), list) and len(plugin_instance.get_raw()) > 0:
            res = False
            for i in plugin_instance.get_raw():
                for f in i:
                    if f in plugin_instance.fields_description:
                        res = True
            self.assertTrue(res)

        self.assertEqual(plugin_instance.get_raw(), plugin_instance.get_export())
        self.assertEqual(plugin_instance.get_stats(), plugin_instance.get_json())
        self.assertEqual(json.loads(plugin_instance.get_stats()), plugin_instance.get_raw())
        if len(plugin_instance.fields_description) > 0:
            # Get first item of the fields_description
            first_field = list(plugin_instance.fields_description.keys())[0]
            self.assertIsInstance(plugin_instance.get_raw_stats_item(first_field), dict)
            self.assertIsInstance(json.loads(plugin_instance.get_stats_item(first_field)), dict)
            self.assertIsInstance(plugin_instance.get_item_info(first_field, 'description'), str)
        # Filter stats
        current_stats = plugin_instance.get_raw()
        if isinstance(plugin_instance.get_raw(), dict):
            current_stats['foo'] = 'bar'
            current_stats = plugin_instance.filter_stats(current_stats)
            self.assertTrue('foo' not in current_stats)
        elif isinstance(plugin_instance.get_raw(), list) and len(plugin_instance.get_raw()) > 0:
            current_stats[0]['foo'] = 'bar'
            current_stats = plugin_instance.filter_stats(current_stats)
            self.assertTrue('foo' not in current_stats[0])

        # Update stats (add second element)
        plugin_instance.update()
        plugin_instance.update_stats_history()
        plugin_instance.update_views()

        # Check stats history
        # Not working on WINDOWS
        if plugin_instance.history_enable() and not WINDOWS:
            if isinstance(plugin_instance.get_raw(), dict):
                first_history_field = plugin_instance.get_items_history_list()[0]['name']
            elif isinstance(plugin_instance.get_raw(), list) and len(plugin_instance.get_raw()) > 0:
                first_history_field = '_'.join(
                    [
                        plugin_instance.get_raw()[0][plugin_instance.get_key()],
                        plugin_instance.get_items_history_list()[0]['name'],
                    ]
                )
            if len(plugin_instance.get_raw()) > 0:
                self.assertEqual(len(plugin_instance.get_raw_history(first_history_field)), 2)
                self.assertGreater(
                    plugin_instance.get_raw_history(first_history_field)[1][0],
                    plugin_instance.get_raw_history(first_history_field)[0][0],
                )
                # Check time
                self.assertEqual(
                    plugin_instance.get_raw_history(first_history_field)[1][0].tzinfo,
                    UTC,
                )

            # Update stats (add third element)
            plugin_instance.update()
            plugin_instance.update_stats_history()
            plugin_instance.update_views()

            if len(plugin_instance.get_raw()) > 0:
                self.assertEqual(len(plugin_instance.get_raw_history(first_history_field)), 3)
                self.assertEqual(len(plugin_instance.get_raw_history(first_history_field, 2)), 2)
                self.assertIsInstance(json.loads(plugin_instance.get_stats_history()), dict)

        # Check views
        self.assertIsInstance(plugin_instance.get_views(), dict)
        self.assertIsInstance(json.loads(plugin_instance.get_json_views()), dict)
        self.assertEqual(json.loads(plugin_instance.get_json_views()), plugin_instance.get_views())
        # Check views history
        # Not working on WINDOWS
        if plugin_instance.history_enable() and not WINDOWS:
            if isinstance(plugin_instance.get_raw(), dict):
                self.assertIsInstance(plugin_instance.get_views(first_history_field), dict)
                self.assertTrue('decoration' in plugin_instance.get_views(first_history_field))
            elif isinstance(plugin_instance.get_raw(), list) and len(plugin_instance.get_raw()) > 0:
                first_history_field = plugin_instance.get_items_history_list()[0]['name']
                first_item = plugin_instance.get_raw()[0][plugin_instance.get_key()]
                self.assertIsInstance(plugin_instance.get_views(item=first_item, key=first_history_field), dict)
                self.assertTrue('decoration' in plugin_instance.get_views(item=first_item, key=first_history_field))

    def test_000_update(self):
        """Update stats (mandatory step for all the stats).

        The update is made twice (for rate computation).
        """
        print('INFO: [TEST_000] Test the stats update function')
        try:
            stats.update()
        except Exception as e:
            print(f'ERROR: Stats update failed: {e}')
            self.assertTrue(False)
        time.sleep(1)
        try:
            stats.update()
        except Exception as e:
            print(f'ERROR: Stats update failed: {e}')
            self.assertTrue(False)

        self.assertTrue(True)

    def test_001_plugins(self):
        """Check mandatory plugins."""
        plugins_to_check = ['system', 'cpu', 'load', 'mem', 'memswap', 'network', 'diskio', 'fs']
        print('INFO: [TEST_001] Check the mandatory plugins list: {}'.format(', '.join(plugins_to_check)))
        plugins_list = stats.getPluginsList()
        for plugin in plugins_to_check:
            self.assertTrue(plugin in plugins_list)

    def test_002_system(self):
        """Check SYSTEM plugin."""
        stats_to_check = ['hostname', 'os_name']
        print('INFO: [TEST_002] Check SYSTEM stats: {}'.format(', '.join(stats_to_check)))
        stats_grab = stats.get_plugin('system').get_raw()
        for stat in stats_to_check:
            # Check that the key exist
            self.assertTrue(stat in stats_grab, msg=f'Cannot find key: {stat}')
        print(f'INFO: SYSTEM stats: {stats_grab}')

    def test_003_cpu(self):
        """Check CPU plugin."""
        stats_to_check = ['system', 'user', 'idle']
        print('INFO: [TEST_003] Check mandatory CPU stats: {}'.format(', '.join(stats_to_check)))
        stats_grab = stats.get_plugin('cpu').get_raw()
        for stat in stats_to_check:
            # Check that the key exist
            self.assertTrue(stat in stats_grab, msg=f'Cannot find key: {stat}')
            # Check that % is > 0 and < 100
            self.assertGreaterEqual(stats_grab[stat], 0)
            self.assertLessEqual(stats_grab[stat], 100)
        print(f'INFO: CPU stats: {stats_grab}')

    @unittest.skipIf(WINDOWS, "Load average not available on Windows")
    def test_004_load(self):
        """Check LOAD plugin."""
        stats_to_check = ['cpucore', 'min1', 'min5', 'min15']
        print('INFO: [TEST_004] Check LOAD stats: {}'.format(', '.join(stats_to_check)))
        stats_grab = stats.get_plugin('load').get_raw()
        for stat in stats_to_check:
            # Check that the key exist
            self.assertTrue(stat in stats_grab, msg=f'Cannot find key: {stat}')
            # Check that % is > 0
            self.assertGreaterEqual(stats_grab[stat], 0)
        print(f'INFO: LOAD stats: {stats_grab}')

    def test_005_mem(self):
        """Check MEM plugin."""
        plugin_name = 'mem'
        stats_to_check = ['available', 'used', 'free', 'total']
        print('INFO: [TEST_005] Check {} stats: {}'.format(plugin_name, ', '.join(stats_to_check)))
        stats_grab = stats.get_plugin('mem').get_raw()
        for stat in stats_to_check:
            # Check that the key exist
            self.assertTrue(stat in stats_grab, msg=f'Cannot find key: {stat}')
            # Check that % is > 0
            self.assertGreaterEqual(stats_grab[stat], 0)
        print(f'INFO: MEM stats: {stats_grab}')

    def test_006_memswap(self):
        """Check MEMSWAP plugin."""
        stats_to_check = ['used', 'free', 'total']
        print('INFO: [TEST_006] Check MEMSWAP stats: {}'.format(', '.join(stats_to_check)))
        stats_grab = stats.get_plugin('memswap').get_raw()
        for stat in stats_to_check:
            # Check that the key exist
            self.assertTrue(stat in stats_grab, msg=f'Cannot find key: {stat}')
            # Check that % is > 0
            self.assertGreaterEqual(stats_grab[stat], 0)
        print(f'INFO: MEMSWAP stats: {stats_grab}')

    def test_007_network(self):
        """Check NETWORK plugin."""
        print('INFO: [TEST_007] Check NETWORK stats')
        stats_grab = stats.get_plugin('network').get_raw()
        self.assertTrue(isinstance(stats_grab, list), msg='Network stats is not a list')
        print(f'INFO: NETWORK stats: {stats_grab}')

    def test_008_diskio(self):
        """Check DISKIO plugin."""
        print('INFO: [TEST_008] Check DISKIO stats')
        stats_grab = stats.get_plugin('diskio').get_raw()
        self.assertTrue(isinstance(stats_grab, list), msg='DiskIO stats is not a list')
        print(f'INFO: diskio stats: {stats_grab}')

    def test_009_fs(self):
        """Check File System plugin."""
        # stats_to_check = [ ]
        print('INFO: [TEST_009] Check FS stats')
        stats_grab = stats.get_plugin('fs').get_raw()
        self.assertTrue(isinstance(stats_grab, list), msg='FileSystem stats is not a list')
        print(f'INFO: FS stats: {stats_grab}')

    def test_010_processes(self):
        """Check Process plugin."""
        # stats_to_check = [ ]
        print('INFO: [TEST_010] Check PROCESS stats')
        stats_grab = stats.get_plugin('processcount').get_raw()
        # total = stats_grab['total']
        self.assertTrue(isinstance(stats_grab, dict), msg='Process count stats is not a dict')
        print(f'INFO: PROCESS count stats: {stats_grab}')
        stats_grab = stats.get_plugin('processlist').get_raw()
        self.assertTrue(isinstance(stats_grab, list), msg='Process count stats is not a list')
        print(f'INFO: PROCESS list stats: {len(stats_grab)} items in the list')
        # Check if number of processes in the list equal counter
        # self.assertEqual(total, len(stats_grab))

    def test_010a_processes_cpu_num(self):
        """Check Process cpu_num (processor) field."""
        print('INFO: [TEST_010a] Check PROCESS cpu_num field')

        # Test 1: Capability detection
        from glances.processes import glances_processes

        self.assertTrue(hasattr(glances_processes, 'disable_cpu_num'))
        print(f'INFO: cpu_num capability - disable_cpu_num={glances_processes.disable_cpu_num}')

        # Test 2: Field in displayed attributes when not disabled
        displayed_attr = glances_processes.get_displayed_attr()
        if glances_processes.disable_cpu_num:
            self.assertNotIn('cpu_num', displayed_attr)
        else:
            self.assertIn('cpu_num', displayed_attr)
        print(f'INFO: cpu_num in displayed_attr: {not glances_processes.disable_cpu_num}')

        # Test 3: Display formatting
        from unittest.mock import Mock

        from glances.plugins.processlist import ProcesslistPlugin

        plugin = ProcesslistPlugin()
        args = Mock()
        args.disable_irix = False
        args.disable_cursor = False

        # Test valid cpu_num value
        result_valid = plugin._get_process_curses_cpu_num({'cpu_num': 5}, False, args)
        self.assertEqual(result_valid.get('msg'), '  5 ')
        self.assertEqual(len(result_valid.get('msg')), 4)

        # Test None value
        result_none = plugin._get_process_curses_cpu_num({'cpu_num': None}, False, args)
        self.assertEqual(result_none.get('msg'), '  - ')

        # Test missing key
        result_missing = plugin._get_process_curses_cpu_num({}, False, args)
        self.assertEqual(result_missing.get('msg'), '  - ')

        print('INFO: cpu_num display formatting tests passed')

    def test_011_folders(self):
        """Check File System plugin."""
        # stats_to_check = [ ]
        print('INFO: [TEST_011] Check FOLDER stats')
        stats_grab = stats.get_plugin('folders').get_raw()
        self.assertTrue(isinstance(stats_grab, list), msg='Folders stats is not a list')
        print(f'INFO: Folders stats: {stats_grab}')

    def test_012_ip(self):
        """Check IP plugin."""
        print('INFO: [TEST_012] Check IP stats')
        stats_grab = stats.get_plugin('ip').get_raw()
        self.assertTrue(isinstance(stats_grab, dict), msg='IP stats is not a dict')
        print(f'INFO: IP stats: {stats_grab}')

    @unittest.skipIf(not LINUX, "IRQs available only on Linux")
    def test_013_irq(self):
        """Check IRQ plugin."""
        print('INFO: [TEST_013] Check IRQ stats')
        stats_grab = stats.get_plugin('irq').get_raw()
        self.assertTrue(isinstance(stats_grab, list), msg='IRQ stats is not a list')
        print(f'INFO: IRQ stats: {stats_grab}')

    @unittest.skipIf(not LINUX, "GPU available only on Linux")
    def test_014_gpu(self):
        """Check GPU plugin."""
        print('INFO: [TEST_014] Check GPU stats')
        stats_grab = stats.get_plugin('gpu').get_raw()
        self.assertTrue(isinstance(stats_grab, list), msg='GPU stats is not a list')
        print(f'INFO: GPU stats: {stats_grab}')

    def test_015_sorted_stats(self):
        """Check sorted stats method."""
        print('INFO: [TEST_015] Check sorted stats method')
        aliases = {
            "key2": "alias11",
            "key5": "alias2",
        }
        unsorted_stats = [
            {"key": "key4"},
            {"key": "key2"},
            {"key": "key5"},
            {"key": "key21"},
            {"key": "key3"},
        ]

        gp = GlancesPluginModel()
        gp.get_key = lambda: "key"
        gp.has_alias = aliases.get
        gp.stats = unsorted_stats

        sorted_stats = gp.sorted_stats()
        self.assertEqual(len(sorted_stats), 5)
        self.assertEqual(sorted_stats[0]["key"], "key5")
        self.assertEqual(sorted_stats[1]["key"], "key2")
        self.assertEqual(sorted_stats[2]["key"], "key3")
        self.assertEqual(sorted_stats[3]["key"], "key4")
        self.assertEqual(sorted_stats[4]["key"], "key21")

    def test_016_subsample(self):
        """Test subsampling function."""
        print('INFO: [TEST_016] Subsampling')
        for l_test in [
            ([1, 2, 3], 4),
            ([1, 2, 3, 4], 4),
            ([1, 2, 3, 4, 5, 6, 7], 4),
            ([1, 2, 3, 4, 5, 6, 7, 8], 4),
            (list(range(1, 800)), 4),
            (list(range(1, 8000)), 800),
        ]:
            l_subsample = subsample(l_test[0], l_test[1])
            self.assertLessEqual(len(l_subsample), l_test[1])

    def test_017_hddsmart(self):
        """Check hard disk SMART data plugin."""
        try:
            from glances.globals import is_admin
        except ImportError:
            print("INFO: [TEST_017] pySMART not found, not running SMART plugin test")
            return

        stat = 'DeviceName'
        print(f'INFO: [TEST_017] Check SMART stats: {stat}')
        stats_grab = stats.get_plugin('smart').get_raw()
        if not is_admin():
            print("INFO: Not admin, SMART list should be empty")
            assert len(stats_grab) == 0
        elif stats_grab == {}:
            print("INFO: Admin but SMART list is empty")
            assert len(stats_grab) == 0
        else:
            print(stats_grab)
            self.assertTrue(stat in stats_grab[0], msg=f'Cannot find key: {stat}')

        print(f'INFO: SMART stats: {stats_grab}')

    def test_017_programs(self):
        """Check Programs plugin."""
        # stats_to_check = [ ]
        print('INFO: [TEST_022] Check PROGRAMS stats')
        stats_grab = stats.get_plugin('programlist').get_raw()
        self.assertTrue(isinstance(stats_grab, list), msg='Programs stats is not a list')
        if stats_grab:
            self.assertTrue(isinstance(stats_grab[0], dict), msg='Programs stats is not a list of dict')
            self.assertTrue('nprocs' in stats_grab[0], msg='No nprocs')

    def test_018_string_value_to_float(self):
        """Check string_value_to_float function"""
        print('INFO: [TEST_018] Check string_value_to_float function')
        self.assertEqual(string_value_to_float('32kB'), 32000.0)
        self.assertEqual(string_value_to_float('32 KB'), 32000.0)
        self.assertEqual(string_value_to_float('15.5MB'), 15500000.0)
        self.assertEqual(string_value_to_float('25.9'), 25.9)
        self.assertEqual(string_value_to_float('12'), 12)
        self.assertEqual(string_value_to_float('--'), None)

    def test_019_events(self):
        """Test events class"""
        print('INFO: [TEST_019] Test events')
        # Init events
        events = GlancesEventsList(max_events=5, min_duration=1, min_interval=3)
        # Minimal event duration not reached
        events.add('WARNING', 'LOAD', 4)
        events.add('CRITICAL', 'LOAD', 5)
        events.add('OK', 'LOAD', 1)
        self.assertEqual(len(events.get()), 0)
        # Minimal event duration LOAD reached
        events.add('WARNING', 'LOAD', 4)
        time.sleep(1)
        events.add('CRITICAL', 'LOAD', 5)
        time.sleep(1)
        events.add('OK', 'LOAD', 1)
        self.assertEqual(len(events.get()), 1)
        self.assertEqual(events.get()[0]['type'], 'LOAD')
        self.assertEqual(events.get()[0]['state'], 'CRITICAL')
        self.assertEqual(events.get()[0]['max'], 5)
        # Minimal event duration CPU reached
        events.add('WARNING', 'CPU', 60)
        time.sleep(1)
        events.add('WARNING', 'CPU', 70)
        time.sleep(1)
        events.add('OK', 'CPU', 10)
        self.assertEqual(len(events.get()), 2)
        self.assertEqual(events.get()[0]['type'], 'CPU')
        self.assertEqual(events.get()[0]['state'], 'WARNING')
        self.assertEqual(events.get()[0]['min'], 60)
        self.assertEqual(events.get()[0]['max'], 70)
        self.assertEqual(events.get()[0]['count'], 2)
        # Minimal event duration CPU reached (again)
        # but time between two events (min_interval) is too short
        # a merge will be done
        time.sleep(0.5)
        events.add('WARNING', 'CPU', 60)
        time.sleep(1)
        events.add('WARNING', 'CPU', 80)
        time.sleep(1)
        events.add('OK', 'CPU', 10)
        self.assertEqual(len(events.get()), 2)
        self.assertEqual(events.get()[0]['type'], 'CPU')
        self.assertEqual(events.get()[0]['state'], 'WARNING')
        self.assertEqual(events.get()[0]['min'], 60)
        self.assertEqual(events.get()[0]['max'], 80)
        self.assertEqual(events.get()[0]['count'], 4)
        # Clean WARNING events
        events.clean()
        self.assertEqual(len(events.get()), 1)

    def test_020_filter(self):
        """Test filter classes"""
        print('INFO: [TEST_020] Test filter')
        gf = GlancesFilter()
        gf.filter = '.*python.*'
        self.assertEqual(gf.filter, '.*python.*')
        self.assertEqual(gf.filter_key, None)
        self.assertTrue(gf.is_filtered({'name': 'python'}))
        self.assertTrue(gf.is_filtered({'name': '/usr/bin/python -m glances'}))
        self.assertFalse(gf.is_filtered({'noname': 'python'}))
        self.assertFalse(gf.is_filtered({'name': 'snake'}))
        gf.filter = 'username:nicolargo'
        self.assertEqual(gf.filter, 'nicolargo')
        self.assertEqual(gf.filter_key, 'username')
        self.assertTrue(gf.is_filtered({'username': 'nicolargo'}))
        self.assertFalse(gf.is_filtered({'username': 'notme'}))
        self.assertFalse(gf.is_filtered({'notuser': 'nicolargo'}))
        gfl = GlancesFilterList()
        gfl.filter = '.*python.*,username:nicolargo'
        self.assertTrue(gfl.is_filtered({'name': 'python is in the place'}))
        self.assertFalse(gfl.is_filtered({'name': 'snake is in the place'}))
        self.assertTrue(gfl.is_filtered({'name': 'snake is in the place', 'username': 'nicolargo'}))
        self.assertFalse(gfl.is_filtered({'name': 'snake is in the place', 'username': 'notme'}))

    def test_021_pretty_date(self):
        """Test pretty_date"""
        print('INFO: [TEST_021] pretty_date')
        self.assertEqual(pretty_date(datetime(2024, 1, 1, 12, 0), datetime(2024, 1, 1, 12, 0)), 'just now')
        self.assertEqual(pretty_date(datetime(2024, 1, 1, 11, 59), datetime(2024, 1, 1, 12, 0)), 'a min')
        self.assertEqual(pretty_date(datetime(2024, 1, 1, 11, 55), datetime(2024, 1, 1, 12, 0)), '5 mins')
        self.assertEqual(pretty_date(datetime(2024, 1, 1, 11, 0), datetime(2024, 1, 1, 12, 0)), 'an hour')
        self.assertEqual(pretty_date(datetime(2024, 1, 1, 0, 0), datetime(2024, 1, 1, 12, 0)), '12 hours')
        self.assertEqual(pretty_date(datetime(2023, 12, 20, 0, 0), datetime(2024, 1, 1, 12, 0)), 'a week')
        self.assertEqual(pretty_date(datetime(2023, 12, 5, 0, 0), datetime(2024, 1, 1, 12, 0)), '3 weeks')
        self.assertEqual(pretty_date(datetime(2023, 12, 1, 0, 0), datetime(2024, 1, 1, 12, 0)), 'a month')
        self.assertEqual(pretty_date(datetime(2023, 6, 1, 0, 0), datetime(2024, 1, 1, 12, 0)), '7 months')
        self.assertEqual(pretty_date(datetime(2023, 1, 1, 0, 0), datetime(2024, 1, 1, 12, 0)), 'an year')
        self.assertEqual(pretty_date(datetime(2020, 1, 1, 0, 0), datetime(2024, 1, 1, 12, 0)), '4 years')

    def test_022_plugin_dag(self):
        """Test Plugin DAG"""
        print('INFO: [TEST_022] Plugins DAG')
        self.assertEqual(get_plugin_dependencies('amps'), ['amps', 'alert'])
        self.assertEqual(get_plugin_dependencies('cpu'), ['cpu', 'core', 'alert'])
        self.assertEqual(get_plugin_dependencies('load'), ['load', 'core', 'alert'])
        self.assertEqual(get_plugin_dependencies('processlist'), ['processlist', 'core', 'processcount', 'alert'])
        self.assertEqual(get_plugin_dependencies('programlist'), ['programlist', 'processcount', 'alert'])
        self.assertEqual(get_plugin_dependencies('quicklook'), ['quicklook', 'fs', 'core', 'load', 'alert'])
        self.assertEqual(get_plugin_dependencies('vms'), ['vms', 'processcount', 'alert'])

    def test_023_get_alert(self):
        """Test get_alert function"""
        print('INFO: [TEST_023] get_alert')
        self.assertEqual(stats.get_plugin('cpu').get_alert(10, minimum=0, maximum=100, header='total'), 'OK_LOG')
        self.assertEqual(stats.get_plugin('cpu').get_alert(65, minimum=0, maximum=100, header='total'), 'CAREFUL_LOG')
        self.assertEqual(stats.get_plugin('cpu').get_alert(75, minimum=0, maximum=100, header='total'), 'WARNING_LOG')
        self.assertEqual(stats.get_plugin('cpu').get_alert(85, minimum=0, maximum=100, header='total'), 'CRITICAL_LOG')

    def test_024_split_esc(self):
        """Test split_esc function"""
        print('INFO: [TEST_024] split_esc')
        self.assertEqual(split_esc(r''), [])
        self.assertEqual(split_esc('\\'), [])
        self.assertEqual(split_esc(r'abcd'), [r'abcd'])
        self.assertEqual(split_esc(r'abcd efg'), [r'abcd', r'efg'])
        self.assertEqual(split_esc('abcd      \n\t\f efg'), [r'abcd', r'efg'])
        self.assertEqual(split_esc(r'abcd\ efg'), [r'abcd efg'])
        self.assertEqual(split_esc(r'', ':'), [''])
        self.assertEqual(split_esc(r'abcd', ':'), [r'abcd'])
        self.assertEqual(split_esc(r'abcd:efg', ':'), [r'abcd', r'efg'])
        self.assertEqual(split_esc(r'abcd\:efg', ':'), [r'abcd:efg'])
        self.assertEqual(split_esc(r'abcd:efg:hijk', ':'), [r'abcd', r'efg', r'hijk'])
        self.assertEqual(split_esc(r'abcd\:efg:hijk', ':'), [r'abcd:efg', r'hijk'])
        self.assertEqual(split_esc(r'abcd\:efg:hijk\:lmnop:qrs', ':', maxsplit=0), [r'abcd\:efg:hijk\:lmnop:qrs'])
        self.assertEqual(split_esc(r'abcd\:efg:hijk\:lmnop:qrs', ':', maxsplit=1), [r'abcd:efg', r'hijk\:lmnop:qrs'])
        self.assertEqual(
            split_esc(r'abcd\:efg:hijk\:lmnop:qrs', ':', maxsplit=10), [r'abcd:efg', r'hijk:lmnop', r'qrs']
        )
        self.assertEqual(split_esc(r'ahellobhelloc', r'hello'), [r'a', r'b', r'c'])
        self.assertEqual(split_esc(r'a\hellobhelloc', r'hello'), [r'ahellob', r'c'])
        self.assertEqual(split_esc(r'ahe\llobhelloc', r'hello'), [r'ahellob', r'c'])

    @unittest.skipIf(not LINUX, "NPU available only on Linux")
    @unittest.skipIf(WINDOWS, "NPU available only on Linux")
    @unittest.skipIf(WSL, "NPU available only on Linux")
    def test_025_npu(self):
        """Check NPU plugin."""
        print('INFO: [TEST_025] Check NPU stats')
        if stats.get_plugin('npu').is_disabled():
            # Disable test if stats is disable in configuration file
            # Related to #3425
            return
        stats_grab = stats.get_plugin('npu').get_raw()
        self.assertTrue(isinstance(stats_grab, list), msg='NPU stats is not a list')
        # Test AMD NPU plugin with test data
        print('INFO: [TEST_025] Check AMD NPU stats with test data')
        stats_amd_npu = NpuPlugin(
            config=test_config, args=test_args, amd_npu_root_folder='./tests-data/plugins/npu/amd'
        )
        stats_amd_npu.update()
        stats_grab = stats_amd_npu.get_raw()
        self.assertTrue(isinstance(stats_grab, list), msg='NPU stats is not a list')
        self.assertEqual(
            stats_grab[0],
            {
                'npu_id': 'amd_1',
                'name': 'AMD NPU (Strix Point)',
                'load': None,
                'freq': 53,
                'freq_current': 800000000,
                'freq_max': 1500000000,
                'mem': None,
                'memory_used': None,
                'memory_total': None,
                'temperature': None,
                'power': None,
            },
        )
        # Test Intel NPU plugin with test data
        print('INFO: [TEST_025] Check Intel NPU stats with test data')
        stats_intel_npu = NpuPlugin(
            config=test_config, args=test_args, intel_npu_root_folder='./tests-data/plugins/npu/intel'
        )
        stats_intel_npu.update()
        stats_grab = stats_intel_npu.get_raw()
        self.assertTrue(isinstance(stats_grab, list), msg='NPU stats is not a list')
        self.assertEqual(
            stats_grab[0],
            {
                'npu_id': 'intel_1',
                'name': 'Intel NPU (Meteor Lake)',
                'load': None,
                'freq': 57,
                'freq_current': 800000000,
                'freq_max': 1400000000,
                'mem': None,
                'memory_used': None,
                'memory_total': None,
                'temperature': 45.0,
                'power': 2.5,
            },
        )
        # Test Rockchip NPU plugin with test data
        print('INFO: [TEST_025] Check Rockchip NPU stats with test data')
        stats_rockchip_npu = NpuPlugin(
            config=test_config, args=test_args, rockchip_npu_root_folder='./tests-data/plugins/npu/rockchip'
        )
        stats_rockchip_npu.update()
        stats_grab = stats_rockchip_npu.get_raw()
        self.assertTrue(isinstance(stats_grab, list), msg='NPU stats is not a list')
        self.assertEqual(
            stats_grab[0],
            {
                'npu_id': 'rockship_1',
                'name': 'Orange Pi 5 Plus',
                'load': 25,
                'freq': 60,
                'freq_current': 600000000,
                'freq_max': 1000000000,
                'mem': None,
                'memory_used': None,
                'memory_total': None,
                'temperature': None,
                'power': None,
            },
        )

    @unittest.skipIf(not LINUX, "MPP available only on Linux")
    @unittest.skipIf(WINDOWS, "MPP available only on Linux")
    @unittest.skipIf(WSL, "MPP available only on Linux")
    def test_026_mpp(self):
        """Check MPP plugin."""
        print('INFO: [TEST_026] Check MPP stats')
        if stats.get_plugin('mpp').is_disabled():
            return
        stats_grab = stats.get_plugin('mpp').get_raw()
        self.assertTrue(isinstance(stats_grab, list), msg='MPP stats is not a list')
        # Test Rockchip MPP plugin with test data
        print('INFO: [TEST_026] Check Rockchip MPP stats with test data')
        stats_rockchip_mpp = MppPlugin(
            config=test_config, args=test_args, rockchip_mpp_root_folder='./tests-data/plugins/mpp/rockchip'
        )
        stats_rockchip_mpp.update()
        stats_grab = stats_rockchip_mpp.get_raw()
        self.assertTrue(isinstance(stats_grab, list), msg='MPP stats is not a list')
        self.assertEqual(len(stats_grab), 3, msg='Expected 3 MPP engines')
        # Build a dict keyed by engine_id for easier assertions
        by_id = {s['engine_id']: s for s in stats_grab}
        self.assertIn('rockchip_rkvenc', by_id)
        self.assertIn('rockchip_rkvdec', by_id)
        self.assertIn('rockchip_jpegd', by_id)
        # Check RKVENC values
        self.assertEqual(by_id['rockchip_rkvenc']['name'], 'RKVENC')
        self.assertEqual(by_id['rockchip_rkvenc']['type'], 'enc')
        self.assertAlmostEqual(by_id['rockchip_rkvenc']['load'], 24.80)
        self.assertAlmostEqual(by_id['rockchip_rkvenc']['utilization'], 24.39)
        self.assertEqual(by_id['rockchip_rkvenc']['sessions'], 1)
        # Check RKVDEC values
        self.assertEqual(by_id['rockchip_rkvdec']['name'], 'RKVDEC')
        self.assertEqual(by_id['rockchip_rkvdec']['type'], 'dec')
        self.assertAlmostEqual(by_id['rockchip_rkvdec']['load'], 28.23)
        self.assertEqual(by_id['rockchip_rkvdec']['sessions'], 1)
        # Check JPEGD values (idle)
        self.assertEqual(by_id['rockchip_jpegd']['name'], 'JPEGD')
        self.assertEqual(by_id['rockchip_jpegd']['type'], 'jpeg')
        self.assertAlmostEqual(by_id['rockchip_jpegd']['load'], 0.0)
        self.assertEqual(by_id['rockchip_jpegd']['sessions'], 0)

    def test_093_auto_unit(self):
        """Test auto_unit classe"""
        print('INFO: [TEST_093] Auto unit')
        self.assertEqual(auto_unit(1.1234), '1.12')
        self.assertEqual(auto_unit(1024), '1024')
        self.assertEqual(auto_unit(1025), '1K')
        self.assertEqual(auto_unit(613421788), '585M')
        self.assertEqual(auto_unit(613421788, low_precision=True), '585M')
        self.assertEqual(auto_unit(5307033647), '4.94G')
        self.assertEqual(auto_unit(5307033647, low_precision=True), '4.9G')
        self.assertEqual(auto_unit(44968414685), '41.9G')
        self.assertEqual(auto_unit(44968414685, low_precision=True), '41.9G')
        self.assertEqual(auto_unit(838471403472), '781G')
        self.assertEqual(auto_unit(838471403472, low_precision=True), '781G')
        self.assertEqual(auto_unit(9683209690677), '8.81T')
        self.assertEqual(auto_unit(9683209690677, low_precision=True), '8.8T')
        self.assertEqual(auto_unit(1073741824), '1024M')
        self.assertEqual(auto_unit(1073741824, low_precision=True), '1024M')
        self.assertEqual(auto_unit(1181116006), '1.10G')
        self.assertEqual(auto_unit(1181116006, low_precision=True), '1.1G')
        # Edge cases:
        m = 2**20
        g = 2**30
        self.assertEqual(auto_unit(9.9 * m), '9.90M')
        self.assertEqual(auto_unit(9.99 * m), '9.99M')
        self.assertEqual(auto_unit(9.991 * m), '9.99M')
        self.assertEqual(auto_unit(9.994 * m), '9.99M')
        self.assertEqual(auto_unit(9.995 * m), '9.99M')
        self.assertEqual(auto_unit(10 * m), '10.0M')
        self.assertEqual(auto_unit(99 * m), '99.0M')
        self.assertEqual(auto_unit(99.9 * m), '99.9M')
        self.assertEqual(auto_unit(99.91 * m), '99.9M')
        self.assertEqual(auto_unit(99.94 * m), '99.9M')
        self.assertEqual(auto_unit(100 * m), '100M')
        # Special cases that used to overflow and misalign columns
        self.assertEqual(auto_unit(9.996 * m), '10.0M')  # was 10.00M
        self.assertEqual(auto_unit(9.999 * m), '10.0M')  # was 10.00M
        self.assertEqual(auto_unit(99.95 * m), '100M')  # was 100.0M
        self.assertEqual(auto_unit(99.96 * m), '100M')  # was 100.0M
        self.assertEqual(auto_unit(99.99 * m), '100M')  # was 100.0M
        self.assertEqual(auto_unit(10 * m - 1), '10.0M')  # was 10.00M
        self.assertEqual(auto_unit(100 * m - 1), '100M')  # was 100.0M
        self.assertEqual(auto_unit(10 * g - 1), '10.0G')  # was 10.00G
        self.assertEqual(auto_unit(100 * g - 1), '100G')  # was 100.0G

    def test_094_thresholds(self):
        """Test thresholds classes"""
        print('INFO: [TEST_094] Thresholds')
        ok = GlancesThresholdOk()
        careful = GlancesThresholdCareful()
        warning = GlancesThresholdWarning()
        critical = GlancesThresholdCritical()
        self.assertTrue(ok < careful)
        self.assertTrue(careful < warning)
        self.assertTrue(warning < critical)
        self.assertFalse(ok > careful)
        self.assertEqual(ok, ok)
        self.assertEqual(str(ok), 'OK')
        thresholds = GlancesThresholds()
        thresholds.add('cpu_percent', 'OK')
        self.assertEqual(thresholds.get(stat_name='cpu_percent').description(), 'OK')

    def test_095_methods(self):
        """Test mandatories methods"""
        print('INFO: [TEST_095] Mandatories methods')
        mandatories_methods = ['reset', 'update']
        plugins_list = stats.getPluginsList()
        for plugin in plugins_list:
            for method in mandatories_methods:
                self.assertTrue(hasattr(stats.get_plugin(plugin), method), msg=f'{plugin} has no method {method}()')

    def test_096_views(self):
        """Test get_views method"""
        print('INFO: [TEST_096] Test views')
        plugins_list = stats.getPluginsList()
        for plugin in plugins_list:
            stats.get_plugin(plugin).get_raw()
            views_grab = stats.get_plugin(plugin).get_views()
            self.assertTrue(isinstance(views_grab, dict), msg=f'{plugin} view is not a dict')

    def test_097_attribute(self):
        """Test GlancesAttribute classes"""
        print('INFO: [TEST_097] Test attribute')
        # GlancesAttribute
        from glances.attribute import GlancesAttribute

        a = GlancesAttribute('a', description='ad', history_max_size=3)
        self.assertEqual(a.name, 'a')
        self.assertEqual(a.description, 'ad')
        a.description = 'adn'
        self.assertEqual(a.description, 'adn')
        a.value = 1
        a.value = 2
        self.assertEqual(len(a.history), 2)
        a.value = 3
        self.assertEqual(len(a.history), 3)
        a.value = 4
        # Check if history_max_size=3 is OK
        self.assertEqual(len(a.history), 3)
        self.assertEqual(a.history_size(), 3)
        self.assertEqual(a.history_len(), 3)
        self.assertEqual(a.history_value()[1], 4)
        self.assertEqual(a.history_mean(nb=3), 4.5)

    def test_098_history(self):
        """Test GlancesHistory classes"""
        print('INFO: [TEST_098] Test history')
        # GlancesHistory
        from glances.history import GlancesHistory

        h = GlancesHistory()
        h.add('a', 1, history_max_size=100)
        h.add('a', 2, history_max_size=100)
        h.add('a', 3, history_max_size=100)
        h.add('b', 10, history_max_size=100)
        h.add('b', 20, history_max_size=100)
        h.add('b', 30, history_max_size=100)
        self.assertEqual(len(h.get()), 2)
        self.assertEqual(len(h.get()['a']), 3)
        h.reset()
        self.assertEqual(len(h.get()), 2)
        self.assertEqual(len(h.get()['a']), 0)

    def test_099_output_bars(self):
        """Test quick look plugin.

        > bar.min_value
        0
        > bar.max_value
        100
        > bar.percent = -1
        > bar.percent
        0
        """
        print('INFO: [TEST_099] Test progress bar')

        bar = Bar(size=1)
        # Percent value can not be lower than min_value
        bar.percent = -1
        self.assertLessEqual(bar.percent, bar.min_value)
        # but... percent value can be higher than max_value
        bar.percent = 101
        self.assertLessEqual(bar.percent, 101)

        # Test display
        bar = Bar(size=50)
        bar.percent = 0
        self.assertEqual(bar.get(), '                                              0.0%')
        bar.percent = 70
        self.assertEqual(bar.get(), '|||||||||||||||||||||||||||||||              70.0%')
        bar.percent = 100
        self.assertEqual(bar.get(), '||||||||||||||||||||||||||||||||||||||||||||  100%')
        bar.percent = 110
        self.assertEqual(bar.get(), '|||||||||||||||||||||||||||||||||||||||||||| >100%')

    # Error in Github Action. Do not remove the comment.
    # def test_100_system_plugin_method(self):
    #     """Test system plugin methods"""
    #     print('INFO: [TEST_100] Test system plugin methods')
    #     self._common_plugin_tests('system')

    def test_101_cpu_plugin_method(self):
        """Test cpu plugin methods"""
        print('INFO: [TEST_100] Test cpu plugin methods')
        self._common_plugin_tests('cpu')

    @unittest.skipIf(WINDOWS, "Load average not available on Windows")
    def test_102_load_plugin_method(self):
        """Test load plugin methods"""
        print('INFO: [TEST_102] Test load plugin methods')
        self._common_plugin_tests('load')

    def test_103_mem_plugin_method(self):
        """Test mem plugin methods"""
        print('INFO: [TEST_103] Test mem plugin methods')
        self._common_plugin_tests('mem')

    def test_104_memswap_plugin_method(self):
        """Test memswap plugin methods"""
        print('INFO: [TEST_104] Test memswap plugin methods')
        self._common_plugin_tests('memswap')

    def test_105_network_plugin_method(self):
        """Test network plugin methods"""
        print('INFO: [TEST_105] Test network plugin methods')
        self._common_plugin_tests('network')

    # Error in Github Action. Do not remove the comment.
    # def test_106_diskio_plugin_method(self):
    #     """Test diskio plugin methods"""
    #     print('INFO: [TEST_106] Test diskio plugin methods')
    #     self._common_plugin_tests('diskio')

    # Before uncommenting this test, please correct deprecation warning
    # ===
    # tests/test_core.py::TestGlances::test_107_fs_plugin_method
    # tests/test_core.py::TestGlances::test_107_fs_plugin_method
    # /python3.14/multiprocessing/popen_fork.py:70: DeprecationWarning:
    # This process (pid=1467579) is multi-threaded, use of fork() may lead to deadlocks in the child.
    #     self.pid = os.fork()
    # -- Docs: https://docs.pytest.org/en/stable/how-to/capture-warnings.html
    # ===
    # def test_107_fs_plugin_method(self):
    #     """Test fs plugin methods"""
    #     print('INFO: [TEST_107] Test fs plugin methods')
    #     self._common_plugin_tests('fs')

    def test_108_fs_zfs_(self):
        """Test zfs functions"""
        print('INFO: [TEST_108] Test zfs functions')
        self.assertTrue(zfs_enable('./tests-data/plugins/fs/zfs'))
        stats = zfs_stats(['./tests-data/plugins/fs/zfs/arcstats'])
        self.assertTrue(isinstance(stats, dict))
        self.assertTrue('arcstats.c_min' in stats)
        self.assertEqual(stats['arcstats.c_min'], 2637352832)
        self.assertTrue('arcstats.size' in stats)
        self.assertEqual(stats['arcstats.size'], 41321273080)

    def test_200_views_hidden(self):
        """Test hide feature"""
        print('INFO: [TEST_200] Test views hidden feature')
        # Test will be done with the diskio plugin, first available interface (read_bytes fields)
        plugin = 'diskio'
        field = 'read_bytes_rate_per_sec'
        plugin_instance = stats.get_plugin(plugin)
        if len(plugin_instance.get_views()) == 0 or not test_config.get_bool_value(plugin, 'hide_zero', False):
            # No diskIO interface, test can not be done
            return
        # Get first disk interface
        key = list(plugin_instance.get_views().keys())[0]
        # Test
        ######
        # Init the stats
        plugin_instance.update()
        raw_stats = plugin_instance.get_raw()
        # Reset the views
        plugin_instance.set_views({})
        # Set field to 0 (should be hidden)
        raw_stats[0][field] = 0
        plugin_instance.set_stats(raw_stats)
        self.assertEqual(plugin_instance.get_raw()[0][field], 0)
        plugin_instance.update_views()
        self.assertTrue(plugin_instance.get_views()[key][field]['hidden'])
        # Set field to 0 (should be hidden)
        raw_stats[0][field] = 0
        plugin_instance.set_stats(raw_stats)
        self.assertEqual(plugin_instance.get_raw()[0][field], 0)
        plugin_instance.update_views()
        self.assertTrue(plugin_instance.get_views()[key][field]['hidden'])
        # Set field to 1 (should not be hidden)
        raw_stats[0][field] = 1
        plugin_instance.set_stats(raw_stats)
        self.assertEqual(plugin_instance.get_raw()[0][field], 1)
        plugin_instance.update_views()
        self.assertFalse(plugin_instance.get_views()[key][field]['hidden'])
        # Set field back to 0 (should not be hidden)
        raw_stats[0][field] = 0
        plugin_instance.set_stats(raw_stats)
        self.assertEqual(plugin_instance.get_raw()[0][field], 0)
        plugin_instance.update_views()
        self.assertFalse(plugin_instance.get_views()[key][field]['hidden'])

    def test_700_mmm_feature_parent_class(self):
        """Test MMM (Min/Max/Mean) feature in parent class."""
        print('INFO: [TEST_700] MMM Feature (Min/Max/Mean) in parent class')

        # Create a test plugin with MMM enabled
        test_fields = {
            'value': {'description': 'A test value', 'unit': 'percent', 'mmm': True},
            'other': {'description': 'A value without MMM', 'unit': 'bytes'},
        }

        class TestMMMPlugin(GlancesPluginModel):
            def __init__(self):
                super().__init__(args=None, config=None, fields_description=test_fields)

        plugin = TestMMMPlugin()

        # Test 1: MMM fields initialized
        self.assertIn('value', plugin._mmm_fields, "value field should be in _mmm_fields")
        self.assertNotIn('other', plugin._mmm_fields, "other field should NOT be in _mmm_fields")

        # Test 2: Auto-generated descriptions exist
        self.assertIn('value_min', plugin.fields_description, "value_min description should be auto-generated")
        self.assertIn('value_max', plugin.fields_description, "value_max description should be auto-generated")
        self.assertIn('value_mean', plugin.fields_description, "value_mean description should be auto-generated")

        # Test 3: Units are inherited
        self.assertEqual(plugin.fields_description['value_min']['unit'], 'percent')
        self.assertEqual(plugin.fields_description['value_max']['unit'], 'percent')
        self.assertEqual(plugin.fields_description['value_mean']['unit'], 'percent')

        # Test 4: MMM field structure is correct
        mmm_info = plugin._mmm_fields['value']
        self.assertIn('values', mmm_info, "values list should exist")
        self.assertIn('min', mmm_info, "min should exist")
        self.assertIn('max', mmm_info, "max should exist")
        self.assertIn('unit', mmm_info, "unit should exist")
        self.assertIsNone(mmm_info['min'], "min should start as None")
        self.assertIsNone(mmm_info['max'], "max should start as None")

    def test_701_mmm_update_functionality(self):
        """Test MMM update functionality."""
        print('INFO: [TEST_701] MMM update functionality')

        test_fields = {'temperature': {'description': 'Temperature in Celsius', 'unit': 'celsius', 'mmm': True}}

        class TestMMMPlugin(GlancesPluginModel):
            def __init__(self):
                super().__init__(args=None, config=None, fields_description=test_fields)

        plugin = TestMMMPlugin()

        # Test 1: First update
        stats = {'temperature': 50.0}
        updated_stats = plugin._update_mmm_fields(stats)

        self.assertIn('temperature_min', updated_stats)
        self.assertIn('temperature_max', updated_stats)
        self.assertIn('temperature_mean', updated_stats)
        self.assertEqual(updated_stats['temperature_min'], 50.0)
        self.assertEqual(updated_stats['temperature_max'], 50.0)
        self.assertEqual(updated_stats['temperature_mean'], 50.0)

        # Test 2: Second update with higher value
        stats = {'temperature': 60.0}
        updated_stats = plugin._update_mmm_fields(stats)

        self.assertEqual(updated_stats['temperature_min'], 50.0, "min should not increase")
        self.assertEqual(updated_stats['temperature_max'], 60.0, "max should increase")
        self.assertLess(50.0, updated_stats['temperature_mean'])
        self.assertLess(updated_stats['temperature_mean'], 60.0)

        # Test 3: Third update with lower value
        stats = {'temperature': 40.0}
        updated_stats = plugin._update_mmm_fields(stats)

        self.assertEqual(updated_stats['temperature_min'], 40.0, "min should decrease")
        self.assertEqual(updated_stats['temperature_max'], 60.0, "max should not decrease")

    def test_702_mmm_mean_calculation(self):
        """Test MMM mean calculation is correct."""
        print('INFO: [TEST_702] MMM mean calculation')

        test_fields = {'value': {'description': 'A test value', 'unit': 'percent', 'mmm': True}}

        class TestMMMPlugin(GlancesPluginModel):
            def __init__(self):
                super().__init__(args=None, config=None, fields_description=test_fields)

        plugin = TestMMMPlugin()

        # Add known values
        test_values = [10.0, 20.0, 30.0, 40.0, 50.0]
        for val in test_values:
            stats = {'value': val}
            updated_stats = plugin._update_mmm_fields(stats)

        # Expected mean of all values
        expected_mean = sum(test_values) / len(test_values)

        self.assertAlmostEqual(updated_stats['value_mean'], expected_mean, places=1)
        # Also verify min and max
        self.assertEqual(updated_stats['value_min'], min(test_values))
        self.assertEqual(updated_stats['value_max'], max(test_values))

    def test_703_mmm_history_limit(self):
        """Test MMM history respects size limit."""
        print('INFO: [TEST_703] MMM history respects size limit')

        test_fields = {'value': {'description': 'A test value', 'unit': 'percent', 'mmm': True}}

        class TestMMMPlugin(GlancesPluginModel):
            def __init__(self):
                super().__init__(args=None, config=None, fields_description=test_fields)

        plugin = TestMMMPlugin()

        # Check default limit
        max_history_size = 28800

        # Add values but only a reasonable number (not the full limit to keep test fast)
        test_iterations = 100
        for i in range(test_iterations):
            stats = {'value': float(i % 100)}
            plugin._update_mmm_fields(stats)

        # History should not exceed limit
        mmm_info = plugin._mmm_fields['value']
        self.assertLessEqual(len(mmm_info['values']), max_history_size)
        # Should have roughly the number we added (100)
        self.assertEqual(len(mmm_info['values']), test_iterations)

    def test_704_mmm_handles_list_of_dicts(self):
        """Test MMM update can handle list of stats dictionaries."""
        print('INFO: [TEST_704] MMM handles list of dicts')

        test_fields = {'utilization': {'description': 'Utilization percent', 'unit': 'percent', 'mmm': True}}

        class TestMMMPlugin(GlancesPluginModel):
            def __init__(self):
                super().__init__(args=None, config=None, fields_description=test_fields)

        plugin = TestMMMPlugin()

        # Create a list of stats
        stats_list = [
            {'name': 'item1', 'utilization': 50.0},
            {'name': 'item2', 'utilization': 60.0},
            {'name': 'item3', 'utilization': 40.0},
        ]

        # Update with list
        updated_list = plugin._update_mmm_fields_on_list(stats_list)

        # Check that each item has MMM fields
        for item in updated_list:
            self.assertIn('utilization_min', item)
            self.assertIn('utilization_max', item)
            self.assertIn('utilization_mean', item)

    def test_705_mmm_ignores_non_numeric_values(self):
        """Test MMM ignores non-numeric values."""
        print('INFO: [TEST_705] MMM ignores non-numeric values')

        test_fields = {'status': {'description': 'Status string', 'unit': 'string', 'mmm': True}}

        class TestMMMPlugin(GlancesPluginModel):
            def __init__(self):
                super().__init__(args=None, config=None, fields_description=test_fields)

        plugin = TestMMMPlugin()

        # Try to update with non-numeric value
        stats = {'status': 'active'}
        updated_stats = plugin._update_mmm_fields(stats)

        # Non-numeric values should not create MMM fields (or they should be None)
        if 'status_min' in updated_stats:
            self.assertIsNone(updated_stats.get('status_min'))

    def test_706_mmm_decorator_integration(self):
        """Test MMM decorator integration with update method."""
        print('INFO: [TEST_706] MMM decorator integration')

        test_fields = {'cpu': {'description': 'CPU usage', 'unit': 'percent', 'mmm': True}}

        class TestMMMPlugin(GlancesPluginModel):
            def __init__(self):
                super().__init__(args=None, config=None, fields_description=test_fields)

            @GlancesPluginModel._manage_mmm
            def test_update(self):
                """Test method with MMM decorator."""
                return {'cpu': 50.5}

        plugin = TestMMMPlugin()

        # Call the decorated method
        result = plugin.test_update()

        # Should have MMM fields
        self.assertIn('cpu_min', result)
        self.assertIn('cpu_max', result)
        self.assertIn('cpu_mean', result)

    def test_707_mmm_with_mem_plugin(self):
        """Test MMM integration with actual MemPlugin."""
        print('INFO: [TEST_707] MMM integration with MemPlugin')

        # Get the mem plugin from stats
        mem_plugin = stats.get_plugin('mem')

        # Verify percent field has MMM enabled
        self.assertTrue(mem_plugin.fields_description['percent'].get('mmm', False))

        # Verify MMM fields are in fields_description
        self.assertIn('percent_min', mem_plugin.fields_description)
        self.assertIn('percent_max', mem_plugin.fields_description)
        self.assertIn('percent_mean', mem_plugin.fields_description)

        # Update and verify stats contain MMM fields
        mem_plugin.update()
        raw_stats = mem_plugin.get_raw()

        self.assertIn('percent', raw_stats)
        self.assertIn('percent_min', raw_stats)
        self.assertIn('percent_max', raw_stats)
        self.assertIn('percent_mean', raw_stats)

        # Verify relationships
        self.assertLessEqual(raw_stats['percent_min'], raw_stats['percent'])
        self.assertGreaterEqual(raw_stats['percent_max'], raw_stats['percent'])
        self.assertLessEqual(raw_stats['percent_min'], raw_stats['percent_mean'])
        self.assertGreaterEqual(raw_stats['percent_max'], raw_stats['percent_mean'])

    # def test_700_secure(self):
    #     """Test secure functions"""
    #     print('INFO: [TEST_700] Secure functions')

    #     if WINDOWS:
    #         self.assertIn(secure_popen('echo TEST'), ['TEST\n', 'TEST\r\n'])
    #         self.assertIn(secure_popen('echo TEST1 && echo TEST2'), ['TEST1\nTEST2\n', 'TEST1\r\nTEST2\r\n'])
    #     else:
    #         self.assertEqual(secure_popen('echo -n TEST'), 'TEST')
    #         self.assertEqual(secure_popen('echo -n TEST1 && echo -n TEST2'), 'TEST1TEST2')
    #         # Make the test failed on Github (AssertionError: '' != 'FOO\n')
    #         # but not on my localLinux computer...
    #         # self.assertEqual(secure_popen('echo FOO | grep FOO'), 'FOO\n')

    def test_999_the_end(self):
        """Free all the stats"""
        print('INFO: [TEST_999] Free the stats')
        stats.end()
        self.assertTrue(True)


if __name__ == '__main__':
    unittest.main()



import json
from unittest.mock import patch

from fastapi.openapi.utils import get_openapi

from glances.main import GlancesMain

# sys.path.append('./glances/outputs')
from glances.outputs.glances_restful_api import GlancesRestfulApi

# Init Glances core
testargs = ["glances", "-C", "./conf/glances.conf"]
with patch('sys.argv', testargs):
    core = GlancesMain()
test_config = core.get_config()
test_args = core.get_args()

app = GlancesRestfulApi(config=test_config, args=test_args)._app

with open('./docs/api/openapi.json', 'w') as f:
    json.dump(
        get_openapi(
            title=app.title,
            version=app.version,
            # Set the OenAPI version
            # It's an hack to make openapi.json compatible with tools like https://editor.swagger.io/
            # Please read https://fastapi.tiangolo.com/reference/fastapi/?h=openapi#fastapi.FastAPI.openapi_version
            openapi_version="3.0.2",
            description=app.description,
            routes=app.routes,
        ),
        f,
    )



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2026 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Intel Extension unit for Glances' NPU plugin.

/sys/class/accel/
└── accel0/
    ├── device -> ../../../0000:00:0b.0/
    │   ├── vendor                    # "0x8086" (Intel)
    │   ├── device                    # Ex: "0x7d1d" (Meteor Lake)
    │   ├── npu_current_frequency_mhz # Ex: "1400"
    │   ├── npu_max_frequency_mhz    # Ex: "1400"
    │   └── hwmon/
    │       └── hwmon2/
    │           ├── temp1_input       # Ex: "45000" (45°C in milli-Celsius)
    │           └── power1_input      # Ex: "2500000" (2.5W in micro-Watts)
    └── dev                           # "261:0"

Current limitation:
- Only one Intel NPU is supported
- Load and Memory stats are not yet implemented
"""

import glob
import os

from glances.logger import logger
from glances.plugins.npu.cards.npu import NPUStats


class IntelNPU:
    """NPU card class."""

    def __init__(self, npu_root_folder: str = '/'):
        """Init Intel NPU card class."""
        self.npu_root_folder = npu_root_folder

        device_paths = glob.glob(os.path.join(self.npu_root_folder, 'sys/class/accel/accel*/device'))
        self.device_folder = device_paths[0] if device_paths else None

        logger.debug(f"Intel NPU device folder: {self.device_folder}")

    def is_available(self) -> bool:
        """Check if Intel NPU is available."""
        return self.device_folder is not None

    def disable(self):
        """Disable Intel GPU class."""
        self.device_folder = None

    def exit(self):
        """Close Intel GPU class."""
        pass

    def get_device_stats(self) -> dict | None:
        """Get Intel GPU stats."""
        if not self.is_available():
            return None

        stats = NPUStats()

        stats.npu_id = 'intel_1'
        stats.freq_current = (
            self._read_file(os.path.join(self.device_folder, 'npu_current_frequency_mhz'), as_int=True) * 1000000
        )  # Convert MHz to Hz
        stats.freq_max = (
            self._read_file(os.path.join(self.device_folder, 'npu_max_frequency_mhz'), as_int=True) * 1000000
        )  # Convert MHz to Hz
        stats.freq = int(stats.freq_current / stats.freq_max * 100)
        stats.name = self._get_device_name(self._read_file(os.path.join(self.device_folder, 'device')))

        # Temperature
        temp_files = glob.glob(os.path.join(self.device_folder, 'hwmon/hwmon*/temp*_input'))
        if temp_files:
            temp_value = self._read_file(temp_files[0], as_int=True)
            if temp_value:
                stats.temperature = temp_value / 1000.0  # Convert milli-Celsius to Celsius

        # Power
        power_files = glob.glob(os.path.join(self.device_folder, 'hwmon/hwmon*/power*_input'))
        if power_files:
            power_value = self._read_file(power_files[0], as_int=True)
            if power_value:
                stats.power = power_value / 1000000.0  # Convert micro-Watts to Watts

        # Return stats as a dictionary
        return stats.__dict__

    def _read_file(self, path: str, as_int: bool = False) -> int | None:
        """Safely read integer from file"""
        try:
            if os.path.exists(path):
                with open(path) as f:
                    if as_int:
                        return int(f.read().strip())
                    return f.read().strip()
        except (OSError, ValueError) as e:
            logger.error(f"NPU - Error reading file {path}: {e}")
        return None

    def _get_device_name(self, device_id: str) -> str:
        """Map Intel device ID to product name"""
        device_map = {
            '0x7d1d': 'Intel NPU (Meteor Lake)',
            '0xa75d': 'Intel NPU (Lunar Lake)',
            '0x643e': 'Intel NPU (Arrow Lake)',
        }
        return device_map.get(device_id, 'Intel NPU')


# End of file



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2026 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Regression tests for GHSA-4h34-v6r8-mmjc (CVE-2026-68520).

Both /api/4/config and /api/4/args serve a sanitised view when Glances runs
without authentication, and both used to filter on key names only:
- `[ip] public_username` was returned in full (the reported leak),
- `[ip] public_api=https://user:pass@host/` leaked the embedded credentials.

/api/4/args had no reachable leak (`-u <login>` forces a password, so the
authenticated view is used), but shared the same design flaw: a hand
maintained key list and no value level check. Both endpoints now go through
the same secure_option() sanitiser.

These tests verify that login names and URL userinfo are redacted, and that
non-sensitive keys sharing the `user` prefix (CPU/load thresholds) are not.
"""

import os
import textwrap
from argparse import Namespace

from glances.config import Config
from glances.outputs.glances_restful_api import GlancesRestfulApi

CONF = """
[ip]
public_api=https://admin:secret123@ipv4.ipleak.net/json/
public_username=myname
public_password=mysecret
public_field=ip

[influxdb]
host=localhost
user=root
password=root

[cpu]
user_careful=50
user_critical=90
user_log=True

[ports]
web_1_url=https://alice:pwd1@blog.example.com
web_2_url=https://bob:pwd2@github.example.com

[passwords]
host_regexp=.*
"""


class _RestfulApiStub(GlancesRestfulApi):
    """GlancesRestfulApi reduced to what _sanitize_args() needs (no server startup)."""

    def __init__(self, args):
        self.args = args


def _sanitize_args(**kwargs):
    """Run GlancesRestfulApi._sanitize_args() against a stub holding only args."""
    defaults = {
        'password': '',
        'username': 'glances',
        'username_used': None,
        'username_prompt': False,
        'password_prompt': False,
        'snmp_user': 'private',
        'snmp_auth': 'password',
        'snmp_community': 'public',
        'conf_file': '/etc/glances/glances.conf',
        'client': None,
        'port': 61208,
        'time': 2.0,
    }
    defaults.update(kwargs)
    return _RestfulApiStub(Namespace(**defaults))._sanitize_args()


def _config(tmp_path):
    """Load a Config from a glances.conf built in tmp_path."""
    conf_file = os.path.join(str(tmp_path), 'glances.conf')
    with open(conf_file, 'w', encoding='utf-8') as f:
        f.write(textwrap.dedent(CONF))
    return Config(config_dir=conf_file)


def test_url_embedded_credentials_are_redacted(tmp_path):
    """A credential bearing URL must not expose its userinfo."""
    secure = _config(tmp_path).as_dict_secure()

    assert secure['ip']['public_api'] == 'https://********@ipv4.ipleak.net/json/'
    assert 'admin' not in secure['ip']['public_api']
    assert 'secret123' not in secure['ip']['public_api']


def test_url_without_credentials_is_untouched(tmp_path):
    """Only the userinfo part is removed, the rest of the value is kept."""
    secure = _config(tmp_path).as_dict_secure()

    assert secure['ip']['public_field'] == 'ip'
    assert secure['influxdb']['host'] == 'localhost'


def test_every_url_of_a_multi_url_value_is_redacted(tmp_path):
    secure = _config(tmp_path).as_dict_secure()

    assert secure['ports']['web_1_url'] == 'https://********@blog.example.com'
    assert secure['ports']['web_2_url'] == 'https://********@github.example.com'


def test_login_names_are_redacted(tmp_path):
    """`username` and `user` are credentials, they must be masked."""
    secure = _config(tmp_path).as_dict_secure()

    assert secure['ip']['public_username'] == '********'
    assert secure['influxdb']['user'] == '********'


def test_passwords_are_still_redacted(tmp_path):
    secure = _config(tmp_path).as_dict_secure()

    assert secure['ip']['public_password'] == '********'
    assert secure['influxdb']['password'] == '********'


def test_passwords_section_is_still_blocked(tmp_path):
    secure = _config(tmp_path).as_dict_secure()

    assert 'passwords' not in secure


def test_cpu_user_thresholds_are_not_redacted(tmp_path):
    """`user_*` thresholds are not credentials, masking them breaks the Web UI."""
    secure = _config(tmp_path).as_dict_secure()

    assert secure['cpu']['user_careful'] == '50'
    assert secure['cpu']['user_critical'] == '90'
    assert secure['cpu']['user_log'] == 'True'


# ---------------------------------------------------------------------------
# /api/4/args
# ---------------------------------------------------------------------------


def test_args_login_is_redacted():
    """Any login name is redacted, not only the `username` key of the hardcoded list."""
    args_json = _sanitize_args(username_used='alice')

    assert args_json['username_used'] == '********'


def test_args_url_embedded_credentials_are_redacted():
    """Defence in depth: no argument carries an URL userinfo today."""
    args_json = _sanitize_args(client='https://admin:secret123@glances.example.com')

    assert args_json['client'] == 'https://********@glances.example.com'


def test_args_hardcoded_sensitive_keys_are_still_redacted():
    args_json = _sanitize_args()

    for key in ('password', 'username', 'snmp_user', 'snmp_auth', 'snmp_community', 'conf_file'):
        assert args_json[key] == '********', key


def test_args_non_string_values_keep_their_type():
    """Booleans cannot carry a credential, redacting them would change the API contract."""
    args_json = _sanitize_args()

    assert args_json['password_prompt'] is False
    assert args_json['username_prompt'] is False
    assert args_json['port'] == 61208
    assert args_json['time'] == 2.0


def test_args_authenticated_view_only_redacts_the_password():
    """With --password set, the authenticated view stays as complete as /api/4/config."""
    args_json = _sanitize_args(password='deadbeef', username_used='alice')

    assert args_json['password'] == '********'
    assert args_json['username_used'] == 'alice'
    assert args_json['conf_file'] == '/etc/glances/glances.conf'



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2022 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Manage the Glances server."""

import fnmatch
import json
import socket
import sys
from base64 import b64decode

from defusedxml import xmlrpc

from glances import __version__
from glances.logger import logger
from glances.processes import glances_processes
from glances.servers_list_dynamic import GlancesAutoDiscoverClient
from glances.stats_server import GlancesStatsServer
from glances.timer import Timer

# Correct issue #1025 by monkey path the xmlrpc lib
xmlrpc.monkey_patch()


class GlancesXMLRPCHandler(xmlrpc.xmlrpc_server.SimpleXMLRPCRequestHandler):
    """Main XML-RPC handler."""

    rpc_paths = ('/RPC2',)

    def end_headers(self):
        # Hack to add a specific header
        # Thk to: https://gist.github.com/rca/4063325
        self.send_my_headers()
        super().end_headers()

    def send_my_headers(self):
        # CORS Access-Control-Allow-Origin (issue #227, GHSA-87qc-fj39-wccr).
        # Per-request reflection against the configured allowlist:
        #   - "*" in the allowlist (default)  -> echo "*"
        #   - request Origin in the allowlist -> reflect it + Vary: Origin
        #   - otherwise                        -> omit ACAO entirely so browsers
        #     block cross-origin JS reads. The XML-RPC call itself still runs;
        #     non-browser clients (curl, custom XML-RPC libs) are unaffected.
        cors_origins = getattr(self.server, 'cors_origins', ['*'])
        if '*' in cors_origins:
            self.send_header("Access-Control-Allow-Origin", "*")
            return
        request_origin = self.headers.get('Origin', '')
        if request_origin and request_origin in cors_origins:
            self.send_header("Access-Control-Allow-Origin", request_origin)
            self.send_header("Vary", "Origin")

    def authenticate(self, headers):
        # auth = headers.get('Authorization')
        try:
            (basic, _, encoded) = headers.get('Authorization').partition(' ')
        except Exception:
            # Client did not ask for authentication
            # If server need it then exit
            return not self.server.isAuth
        else:
            # Client authentication
            (basic, _, encoded) = headers.get('Authorization').partition(' ')
            assert basic == 'Basic', 'Only basic authentication supported'
            # Encoded portion of the header is a string
            # Need to convert to byte-string
            encoded_byte_string = encoded.encode()
            # Decode base64 byte string to a decoded byte string
            decoded_bytes = b64decode(encoded_byte_string)
            # Convert from byte string to a regular string
            decoded_string = decoded_bytes.decode()
            # Get the username and password from the string
            (username, _, password) = decoded_string.partition(':')
            # Check that username and password match internal global dictionary
            return self.check_user(username, password)

    def check_user(self, username, password):
        # Check username and password in the dictionary
        if username in self.server.user_dict:
            from glances.password import GlancesPassword

            # TODO: config is not taken into account: it may be a problem ?
            pwd = GlancesPassword(username=username, config=None)
            return pwd.check_password(self.server.user_dict[username], password)
        return False

    def parse_request(self):
        if not xmlrpc.xmlrpc_server.SimpleXMLRPCRequestHandler.parse_request(self):
            return False
        # Host header validation (DNS rebinding protection, GHSA-w856-8p3r-p338).
        # Runs before authentication so a spoofed Host is rejected regardless of
        # credentials — avoids leaking valid hostnames via auth-error differentials.
        allowed_hosts = getattr(self.server, 'allowed_hosts', None)
        if allowed_hosts:
            host = self.headers.get('Host', '').split(':')[0]
            if not any(fnmatch.fnmatchcase(host, pat) for pat in allowed_hosts):
                self.send_error(400, 'Bad Request: invalid Host header')
                return False
        if self.authenticate(self.headers):
            return True
        # if authentication fails, tell the client
        self.send_error(401, 'Authentication failed')
        return False

    def log_message(self, log_format, *args):
        # No message displayed on the server side
        pass


class GlancesXMLRPCServer(xmlrpc.xmlrpc_server.SimpleXMLRPCServer):
    """Init a SimpleXMLRPCServer instance (IPv6-ready)."""

    finished = False

    def __init__(self, bind_address, bind_port=61209, requestHandler=GlancesXMLRPCHandler, config=None):
        self.bind_address = bind_address
        self.bind_port = bind_port
        self.config = config

        # CORS origins (GHSA-87qc-fj39-wccr / CVE-2026-46608).
        # Stored as a list; the handler reflects the request Origin against it
        # per request. Default ["*"] preserves backward compatibility for the
        # majority of users who run on a private network without configuration.
        if config is not None:
            self.cors_origins = config.get_list_value('outputs', 'cors_origins', default=["*"])
        else:
            self.cors_origins = ["*"]

        # DNS rebinding protection (GHSA-w856-8p3r-p338 / CVE-2026-46611).
        # When xmlrpc_allowed_hosts is set, the handler rejects requests whose
        # Host header does not match any of the listed patterns. Default is
        # None (permissive) for backward compatibility — a startup warning is
        # emitted by GlancesServer when the allowlist is empty.
        if config is not None:
            self.allowed_hosts = config.get_list_value('outputs', 'xmlrpc_allowed_hosts', default=None)
        else:
            self.allowed_hosts = None

        try:
            self.address_family = socket.getaddrinfo(bind_address, bind_port)[0][0]
        except OSError as e:
            logger.error(f"Couldn't open socket: {e}")
            sys.exit(1)

        super().__init__((bind_address, bind_port), requestHandler)

    def end(self):
        """Stop the server"""
        self.server_close()
        self.finished = True

    def serve_forever(self):
        """Main loop"""
        while not self.finished:
            self.handle_request()


class GlancesInstance:
    """All the methods of this class are published as XML-RPC methods."""

    def __init__(self, config=None, args=None):
        # Init stats
        self.stats = GlancesStatsServer(config=config, args=args)

        # Initial update
        self.stats.update()

        # cached_time is the minimum time interval between stats updates
        # i.e. XML/RPC calls will not retrieve updated info until the time
        # since last update is passed (will retrieve old cached info instead)
        self.timer = Timer(0)
        self.cached_time = args.cached_time

    def __update__(self):
        # Never update more than 1 time per cached_time
        if self.timer.finished():
            self.stats.update()
            self.timer = Timer(self.cached_time)

    def init(self):
        # Return the Glances version
        return __version__

    def getAll(self):
        # Update and return all the stats
        self.__update__()
        return json.dumps(self.stats.getAll())

    def getAllPlugins(self):
        # Return the plugins list
        return json.dumps(self.stats.getPluginsList())

    def getAllLimits(self):
        # Return all the plugins limits
        return json.dumps(self.stats.getAllLimitsAsDict())

    def getAllViews(self):
        # Return all the plugins views
        return json.dumps(self.stats.getAllViewsAsDict())

    def getPlugin(self, plugin):
        # Update and return the plugin stat
        self.__update__()
        return json.dumps(self.stats.get_plugin(plugin).get_raw())

    def getPluginView(self, plugin):
        # Update and return the plugin view
        return json.dumps(self.stats.get_plugin(plugin).get_views())


class GlancesServer:
    """This class creates and manages the TCP server."""

    def __init__(self, requestHandler=GlancesXMLRPCHandler, config=None, args=None):
        # Args
        self.args = args

        # Set the args for the glances_processes instance
        glances_processes.set_args(args)

        # Init the XML RPC server
        try:
            self.server = GlancesXMLRPCServer(args.bind_address, args.port, requestHandler, config=config)
        except Exception as e:
            logger.critical(f"Cannot start Glances server: {e}")
            sys.exit(2)
        else:
            print(f'Glances XML-RPC server is running on {args.bind_address}:{args.port}')

        # The users dict
        # username / password couple
        # By default, no auth is needed
        self.server.user_dict = {}
        self.server.isAuth = False

        # Warn if running unauthenticated with wildcard CORS
        if args.password == "" and "*" in self.server.cors_origins:
            print(
                "WARNING: XML-RPC server is running without authentication and with CORS Allow-Origin: *.\n"
                "         Mitigations: set a password (-P/--password) and/or restrict\n"
                "         cors_origins in glances.conf [outputs] section."
            )
            logger.warning(
                "XML-RPC server is running without authentication and with CORS Access-Control-Allow-Origin: *. "
            )

        # Warn if running without Host header validation (DNS rebinding protection)
        if not getattr(self.server, 'allowed_hosts', None):
            print(
                "WARNING: Glances XML-RPC server is running without Host header validation.\n"
                "         DNS rebinding attacks may allow untrusted pages to read the XML-RPC API.\n"
                "         Set xmlrpc_allowed_hosts in glances.conf [outputs] to restrict access:\n"
                "           xmlrpc_allowed_hosts=localhost,127.0.0.1,<your-hostname>"
            )
            logger.warning("XML-RPC server is running without Host header validation (DNS rebinding protection)")

        # Register functions
        self.server.register_introspection_functions()
        self.server.register_instance(GlancesInstance(config, args))

        if not self.args.disable_autodiscover:
            # Note: The Zeroconf service name will be based on the hostname
            # Correct issue: Zeroconf problem with zeroconf service name #889
            logger.info('Autodiscover is enabled with service name {}'.format(socket.gethostname().split('.', 1)[0]))
            self.autodiscover_client = GlancesAutoDiscoverClient(socket.gethostname().split('.', 1)[0], args)
        else:
            logger.info("Glances autodiscover announce is disabled")

    def add_user(self, username, password):
        """Add an user to the dictionary."""
        self.server.user_dict[username] = password
        self.server.isAuth = True

    def serve_forever(self):
        """Call the main loop."""
        # Set the server login/password (if -P/--password tag)
        if self.args.password != "":
            self.add_user(self.args.username, self.args.password)
        # Serve forever
        self.server.serve_forever()

    def end(self):
        """End of the Glances server session."""
        if not self.args.disable_autodiscover:
            self.autodiscover_client.close()
        self.server.end()



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""MCP (Model Context Protocol) server interface class."""

import json
import logging

from glances.globals import json_dumps
from glances.logger import logger

try:
    from mcp.server.fastmcp import FastMCP
    from mcp.server.sse import TransportSecuritySettings

    MCP_AVAILABLE = True
    logging.getLogger("mcp").setLevel(logging.WARNING)
except ImportError:
    MCP_AVAILABLE = False
    FastMCP = None
    TransportSecuritySettings = None


class GlancesMcpServer:
    """MCP server exposing Glances system monitoring data as resources and prompts.

    Resources (read-only data):
        glances://plugins                   - List of active plugins
        glances://stats                     - All plugins' current statistics
        glances://stats/{plugin}            - Current statistics for one plugin
        glances://stats/{plugin}/history    - Historical time-series for one plugin
        glances://limits                    - Alert thresholds for all plugins
        glances://limits/{plugin}           - Alert thresholds for one plugin

    Prompts (analysis templates):
        system_health_summary   - Overall CPU / memory / disk / network health report
        alert_analysis          - Analysis of active alerts with optional severity filter
        top_processes_report    - Report on the most CPU-intensive processes
        storage_health          - Disk usage and I/O statistics analysis

    Usage:
        mcp_server = GlancesMcpServer(stats, args, config)
        app.mount("/mcp", mcp_server.get_asgi_app())

    Authentication:
        The MCP endpoint inherits the authentication policy applied at the FastAPI
        router/middleware level (see glances_restful_api.py).  No authentication
        logic lives inside this class.
    """

    MCP_DEFAULT_PATH = "/mcp"

    def __init__(self, stats, args, config):
        """Initialize the MCP server.

        Args:
            stats: GlancesStats instance.  May be None at construction time if the
                   stats manager has not been created yet; call set_stats() before
                   the server receives its first request.
            args:  Parsed command-line arguments (argparse.Namespace).
            config: Glances configuration object (GlancesConfig).
        """
        self._stats = stats
        self._args = args
        self._config = config

        if not MCP_AVAILABLE:
            raise RuntimeError(
                "The 'mcp' package is required for MCP support. Install it with: pip install 'glances[mcp]'"
            )

        self._load_config(config)

        self._mcp = FastMCP(
            "Glances",
            instructions=(
                "Glances is a cross-platform system monitoring tool. "
                "Use resources to access real-time system metrics "
                "(CPU, memory, disk, network, processes, containers, sensors, …). "
                "Use prompts to generate structured analyses of the current system state."
            ),
            transport_security=self._build_transport_security(),
        )

        self._setup_resources()
        self._setup_prompts()
        logger.debug("GlancesMcpServer: resources and prompts registered")

    # ------------------------------------------------------------------
    # Public helpers
    # ------------------------------------------------------------------

    def set_stats(self, stats):
        """Update the stats manager reference.

        Call this method after the GlancesStats object is created when it was
        not yet available at construction time.
        """
        self._stats = stats

    def get_asgi_app(self, mount_path: str = MCP_DEFAULT_PATH):
        """Return the SSE ASGI application ready to be mounted into FastAPI.

        Args:
            mount_path: Informational only — the URL path where this app is mounted
                        in the parent FastAPI instance (used for the log message).
                        Do NOT forward this to FastMCP.sse_app(): when Starlette
                        mounts a sub-application it sets scope['root_path'] to the
                        mount prefix automatically, and SseServerTransport uses that
                        value to build the correct endpoint URL for clients.
                        Passing mount_path again would cause the prefix to appear
                        twice (e.g. /mcp/mcp/messages/).

        Returns:
            A Starlette ASGI application.
        """
        logger.debug(f"MCP server (SSE transport) mounted at {mount_path} with allowed hosts: {self.mcp_allowed_hosts}")
        # Call sse_app() without mount_path so FastMCP keeps its default '/'.
        # Starlette will prepend scope['root_path'] (= mount_path) at runtime,
        # producing the correct absolute endpoint URL for clients.
        # transport_security is already configured on self._mcp at construction time.
        return self._mcp.sse_app()

    # ------------------------------------------------------------------
    # Internal helpers
    # ------------------------------------------------------------------

    def _load_config(self, config):
        # MCP allowed hosts for DNS rebinding protection
        # Default: loopback only (safe for existing deployments)
        self.mcp_allowed_hosts = ["localhost", "127.0.0.1"]
        if config is not None and config.has_section('outputs'):
            raw_mcp_hosts = config.get_value("outputs", "mcp_allowed_hosts", default=None)
            if raw_mcp_hosts == "*":
                self.mcp_allowed_hosts = ["*"]
                logger.warning(
                    "MCP mcp_allowed_hosts is set to wildcard (*). Ensure the server is behind a trusted reverse proxy."
                )
            elif raw_mcp_hosts:
                self.mcp_allowed_hosts = [h.strip() for h in raw_mcp_hosts.split(",") if h.strip()]
        logger.debug(f"MCP allowed hosts: {self.mcp_allowed_hosts}")

    def _build_transport_security(self):
        """Return TransportSecuritySettings derived from mcp_allowed_hosts.

        Wildcard ("*") disables DNS rebinding protection entirely.
        Otherwise each hostname is converted to a "host:*" pattern (any port)
        and paired with http/https origin patterns.
        """
        if self.mcp_allowed_hosts == ["*"]:
            return TransportSecuritySettings(enable_dns_rebinding_protection=False)

        # Append :* to plain hostnames so the middleware matches any port.
        allowed_hosts = [h if ":" in h else f"{h}:*" for h in self.mcp_allowed_hosts]
        allowed_origins = [f"http://{h}" for h in allowed_hosts] + [f"https://{h}" for h in allowed_hosts]
        return TransportSecuritySettings(
            enable_dns_rebinding_protection=True,
            allowed_hosts=allowed_hosts,
            allowed_origins=allowed_origins,
        )

    def _get_stats(self):
        """Return the stats manager, raising RuntimeError if not yet set."""
        if self._stats is None:
            raise RuntimeError(
                "GlancesMcpServer: stats manager is not yet initialized. Call set_stats() before serving MCP requests."
            )
        return self._stats

    def _serialize(self, data) -> str:
        """Return a UTF-8 JSON string using Glances' custom serializer.

        Glances' json_dumps handles non-standard numeric types and datetime
        objects that the stdlib json module cannot encode out of the box.
        """
        return json_dumps(data).decode("utf-8")

    # ------------------------------------------------------------------
    # Resources
    # ------------------------------------------------------------------

    def _setup_resources(self):
        """Declare all MCP resources on the FastMCP instance."""
        mcp = self._mcp
        server = self  # captured by closures below

        # ---- plugins list ------------------------------------------------

        @mcp.resource(
            "glances://plugins",
            name="plugins_list",
            description=(
                "List of all active monitoring plugin names available in this "
                "Glances instance (e.g. 'cpu', 'mem', 'network', 'processlist', …)."
            ),
            mime_type="application/json",
        )
        def plugins_list() -> str:
            return server._serialize(server._get_stats().getPluginsList())

        # ---- all stats ---------------------------------------------------

        @mcp.resource(
            "glances://stats",
            name="all_stats",
            description=(
                "Current statistics from every active monitoring plugin. Returns a JSON object keyed by plugin name."
            ),
            mime_type="application/json",
        )
        def all_stats() -> str:
            return server._serialize(server._get_stats().getAllAsDict())

        # ---- per-plugin stats (resource template) ------------------------

        @mcp.resource(
            "glances://stats/{plugin}",
            name="plugin_stats",
            description=(
                "Current statistics for a specific monitoring plugin. "
                "Fetch glances://plugins first to discover available plugin names."
            ),
            mime_type="application/json",
        )
        def plugin_stats(plugin: str) -> str:
            stats = server._get_stats()
            plugin_obj = stats.get_plugin(plugin)
            if plugin_obj is None:
                available = stats.getPluginsList()
                raise ValueError(f"Plugin '{plugin}' not found. Available plugins: {available}")
            return server._serialize(plugin_obj.get_raw())

        # ---- per-plugin history (resource template) ----------------------

        @mcp.resource(
            "glances://stats/{plugin}/history",
            name="plugin_history",
            description=(
                "Historical time-series data for a specific monitoring plugin. "
                "Returns a dict of field→list pairs, most-recent value last."
            ),
            mime_type="application/json",
        )
        def plugin_history(plugin: str) -> str:
            stats = server._get_stats()
            plugin_obj = stats.get_plugin(plugin)
            if plugin_obj is None:
                raise ValueError(f"Plugin '{plugin}' not found")
            # nb=0 → return all available history points
            return server._serialize(plugin_obj.get_raw_history(item=None, nb=0))

        # ---- all limits --------------------------------------------------

        @mcp.resource(
            "glances://limits",
            name="all_limits",
            description=(
                "Warning and critical alert thresholds for all monitoring plugins. "
                "Returns a JSON object keyed by plugin name."
            ),
            mime_type="application/json",
        )
        def all_limits() -> str:
            return server._serialize(server._get_stats().getAllLimitsAsDict())

        # ---- per-plugin limits (resource template) -----------------------

        @mcp.resource(
            "glances://limits/{plugin}",
            name="plugin_limits",
            description=("Warning and critical alert thresholds for a specific monitoring plugin."),
            mime_type="application/json",
        )
        def plugin_limits(plugin: str) -> str:
            stats = server._get_stats()
            plugin_obj = stats.get_plugin(plugin)
            if plugin_obj is None:
                raise ValueError(f"Plugin '{plugin}' not found")
            return server._serialize(plugin_obj.get_limits())

    # ------------------------------------------------------------------
    # Prompts
    # ------------------------------------------------------------------

    def _setup_prompts(self):
        """Declare all MCP prompt templates on the FastMCP instance."""
        mcp = self._mcp
        server = self  # captured by closures below

        # ---- system health summary ---------------------------------------

        @mcp.prompt(
            name="system_health_summary",
            description=(
                "Generate a comprehensive system health report covering CPU, "
                "memory, swap, load average, filesystems, and network interfaces."
            ),
        )
        def system_health_summary() -> str:
            stats = server._get_stats()
            data = {
                plugin: stats.get_plugin(plugin).get_raw()
                for plugin in ("cpu", "mem", "memswap", "load", "fs", "network")
                if stats.get_plugin(plugin) is not None
            }
            return (
                "You are a system administrator assistant. "
                "Analyze the following real-time system monitoring data collected by Glances "
                "and produce a concise health report.\n"
                "Highlight any metrics that exceed warning or critical thresholds, "
                "identify potential bottlenecks, and suggest remediation steps where needed.\n\n"
                f"System data (JSON):\n{json.dumps(data, default=str, indent=2)}"
            )

        # ---- alert analysis ----------------------------------------------

        @mcp.prompt(
            name="alert_analysis",
            description=(
                "Analyze active Glances system alerts and warnings. "
                "Accepts an optional 'level' parameter: 'warning', 'critical', or 'all' (default)."
            ),
        )
        def alert_analysis(level: str = "all") -> str:
            stats = server._get_stats()
            alert_plugin = stats.get_plugin("alert")
            data = alert_plugin.get_raw() if alert_plugin is not None else []
            return (
                "You are a system administrator assistant. "
                f"Analyze the following Glances system alerts (severity filter: '{level}').\n"
                "For each alert, explain the likely cause, the risk level, "
                "and the recommended corrective action.\n\n"
                f"Active alerts (JSON):\n{json.dumps(data, default=str, indent=2)}"
            )

        # ---- top processes report ----------------------------------------

        @mcp.prompt(
            name="top_processes_report",
            description=(
                "Report on the most resource-consuming processes currently running. "
                "Accepts an optional 'nb' parameter for the number of processes (default: 10)."
            ),
        )
        def top_processes_report(nb: int = 10) -> str:
            stats = server._get_stats()
            proc_plugin = stats.get_plugin("processlist")
            processes: list = []
            if proc_plugin is not None:
                all_procs = proc_plugin.get_raw() or []
                processes = sorted(
                    all_procs,
                    key=lambda p: p.get("cpu_percent", 0),
                    reverse=True,
                )[:nb]
            return (
                "You are a system administrator assistant. "
                f"Analyze the following top {nb} processes sorted by CPU usage, "
                "as reported by Glances.\n"
                "Identify processes that may be causing performance issues, "
                "explain what each process likely does, "
                "and suggest optimization steps where appropriate.\n\n"
                f"Top processes (JSON):\n{json.dumps(processes, default=str, indent=2)}"
            )

        # ---- storage health ----------------------------------------------

        @mcp.prompt(
            name="storage_health",
            description=(
                "Analyze disk usage and I/O statistics to assess storage health and identify potential issues."
            ),
        )
        def storage_health() -> str:
            stats = server._get_stats()
            data = {
                plugin: stats.get_plugin(plugin).get_raw()
                for plugin in ("fs", "diskio")
                if stats.get_plugin(plugin) is not None
            }
            return (
                "You are a system administrator assistant. "
                "Analyze the following disk and filesystem statistics collected by Glances.\n"
                "Identify filesystems that are critically full, "
                "highlight unusual I/O patterns, "
                "and provide actionable recommendations.\n\n"
                f"Storage data (JSON):\n{json.dumps(data, default=str, indent=2)}"
            )



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2026 Christian Rishøj <christian@rishoj.net>
#
# SPDX-License-Identifier: LGPL-3.0-only

"""LXD Extension unit for Glances' Containers plugin."""

import threading
import time
from datetime import datetime
from typing import Any

import psutil

from glances.globals import nativestr, pretty_date
from glances.logger import logger

# pylxd library (optional and Linux-only)
# https://github.com/canonical/pylxd
try:
    from pylxd import Client as LxdClient
except Exception as e:
    disable_plugin_lxd = True
    logger.warning(f"Error loading LXD deps Lib. LXD feature in the Containers plugin is disabled ({e})")
else:
    disable_plugin_lxd = False


class LxdStatsFetcher:
    """Fetch stats for a single LXD instance by polling its state."""

    def __init__(self, instance, poll_interval=2):
        self._instance = instance
        self._poll_interval = poll_interval

        # Store previous stats for rate calculations
        self._old_computed_stats = {}
        self._last_stats_computed_time = time.time()

        # Latest polled state
        self._state = None
        self._state_lock = threading.Lock()

        # Polling thread
        self._stopper = threading.Event()
        self._thread = threading.Thread(target=self._poll_loop, daemon=True)
        self._thread.start()

    def _poll_loop(self):
        """Poll instance.state() in a background thread."""
        while not self._stopper.is_set():
            try:
                state = self._instance.state()
                with self._state_lock:
                    self._state = state
            except Exception as e:
                logger.debug(f"containers (LXD) Instance({self._instance.name}): Failed to poll state ({e})")
            self._stopper.wait(self._poll_interval)

    def stop(self):
        self._stopper.set()

    @property
    def activity_stats(self) -> dict[str, dict[str, Any]]:
        """Compute activity stats from the latest polled state."""
        computed = self._compute_activity_stats()
        self._old_computed_stats = computed
        self._last_stats_computed_time = time.time()
        return computed

    @property
    def time_since_update(self) -> float:
        return max(1, time.time() - self._last_stats_computed_time)

    def _compute_activity_stats(self) -> dict[str, dict[str, Any]]:
        with self._state_lock:
            state = self._state

        if state is None:
            return {"cpu": {}, "memory": {}, "io": {}, "network": {}}

        return {
            "cpu": self._get_cpu_stats(state),
            "memory": self._get_memory_stats(state),
            "io": self._get_io_stats(state),
            "network": self._get_network_stats(state),
        }

    def _get_cpu_stats(self, state) -> dict[str, Any]:
        """Return CPU usage stats.

        LXD reports cumulative CPU time in nanoseconds.
        We compute a percentage from the delta between two polls.
        """
        stats = {}
        try:
            cumulative_cpu_ns = state.cpu["usage"]
            stats["cumulative_ns"] = cumulative_cpu_ns

            old_cpu = self._old_computed_stats.get("cpu", {})
            if "cumulative_ns" in old_cpu:
                delta_ns = cumulative_cpu_ns - old_cpu["cumulative_ns"]
                delta_seconds = self.time_since_update
                # Convert ns delta to a percentage (assuming 1 core = 100%)
                stats["total"] = (delta_ns / (delta_seconds * 1e9)) * 100.0
            else:
                stats["total"] = 0.0
        except (KeyError, TypeError, ZeroDivisionError) as e:
            logger.debug(f"containers (LXD) Instance({self._instance.name}): Can't grab CPU stats ({e})")

        try:
            limit = self._instance.expanded_config.get('limits.cpu')
            if limit:
                stats['limit'] = float(limit)
            else:
                stats['limit'] = float(psutil.cpu_count())
        except (ValueError, TypeError):
            stats['limit'] = float(psutil.cpu_count())

        return stats

    def _get_memory_stats(self, state) -> dict[str, Any]:
        """Return memory usage stats.

        LXD reports memory values in bytes.
        'total' is 0 when unlimited; fall back to usage_peak.
        """
        stats = {}
        try:
            stats["usage"] = state.memory["usage"]
            mem_total = state.memory.get("total", 0)
            if mem_total > 0:
                stats["limit"] = mem_total
            else:
                stats["limit"] = state.memory.get("usage_peak", state.memory["usage"])
            stats["inactive_file"] = 0
        except (KeyError, TypeError) as e:
            logger.debug(f"containers (LXD) Instance({self._instance.name}): Can't grab MEM stats ({e})")
        return stats

    def _get_io_stats(self, state) -> dict[str, Any]:
        """Return disk IO stats.

        LXD only exposes cumulative disk usage per device, not separate read/write.
        """
        stats = {}
        try:
            if state.disk and "root" in state.disk:
                cumulative_disk = state.disk["root"].get("usage", 0)
                stats["cumulative_ior"] = cumulative_disk
                stats["cumulative_iow"] = 0

                old_io = self._old_computed_stats.get("io", {})
                if "cumulative_ior" in old_io:
                    stats["time_since_update"] = round(self.time_since_update)
                    stats["ior"] = max(0, cumulative_disk - old_io["cumulative_ior"])
                    stats["iow"] = 0
        except (KeyError, TypeError) as e:
            logger.debug(f"containers (LXD) Instance({self._instance.name}): Can't grab IO stats ({e})")
        return stats

    def _get_network_stats(self, state) -> dict[str, Any]:
        """Return network usage stats.

        Sums counters across all non-loopback interfaces.
        """
        stats = {}
        try:
            if state.network:
                cumulative_rx = 0
                cumulative_tx = 0
                for iface in state.network.values():
                    if iface.get("type") == "loopback":
                        continue
                    counters = iface.get("counters", {})
                    cumulative_rx += counters.get("bytes_received", 0)
                    cumulative_tx += counters.get("bytes_sent", 0)

                stats["cumulative_rx"] = cumulative_rx
                stats["cumulative_tx"] = cumulative_tx

                old_net = self._old_computed_stats.get("network", {})
                if "cumulative_rx" in old_net:
                    stats["time_since_update"] = round(self.time_since_update)
                    stats["rx"] = max(0, cumulative_rx - old_net["cumulative_rx"])
                    stats["tx"] = max(0, cumulative_tx - old_net["cumulative_tx"])
        except (KeyError, TypeError) as e:
            logger.debug(f"containers (LXD) Instance({self._instance.name}): Can't grab NET stats ({e})")
        return stats


class LxdExtension:
    """Glances' Containers Plugin's LXD Extension unit"""

    CONTAINER_ACTIVE_STATUS = ['Running']

    def __init__(self, endpoint=None, poll_interval=2):
        self.disable = disable_plugin_lxd
        if self.disable:
            raise Exception("Missing libs required to run LXD Extension (Containers)")

        self.display_error = True
        self.client = None
        self.ext_name = "containers (LXD)"
        self.endpoint = endpoint
        self.poll_interval = poll_interval
        self.stats_fetchers = {}
        self.local_node = None

        self.connect()

    def connect(self) -> None:
        """Connect to the LXD server."""
        try:
            if self.endpoint:
                self.client = LxdClient(endpoint=self.endpoint)
            else:
                self.client = LxdClient()
            # Verify connectivity
            self.client.has_api_extension('instances')
            # In a cluster, determine local member name so we can filter to instances
            # running on this node. On a standalone host, server_name is still set but
            # instance.location is empty, so we must gate on server_clustered.
            try:
                env = self.client.host_info.get('environment', {})
                if env.get('server_clustered'):
                    self.local_node = env.get('server_name')
            except Exception:
                self.local_node = None
        except Exception as e:
            logger.error(f"{self.ext_name} plugin - Can't connect to LXD ({e})")
            self.client = None
            self.disable = True

    def stop(self) -> None:
        for t in self.stats_fetchers.values():
            t.stop()

    def update(self, all_tag) -> tuple[dict, list[dict[str, Any]]]:
        """Update LXD stats using the input method."""
        if not self.client or self.disable:
            return {}, []

        # List instances
        try:
            # Issue #3559: recursion=1 fetches every instance's config (status, devices,
            # created_at, ...) in a single API call. Without it, pylxd creates name-only
            # instances and each attribute access below (i.status, expanded_devices, ...)
            # lazily triggers one synchronous GET per instance, i.e. O(N) calls per cycle.
            # State (CPU/MEM/NET/IO) still comes from the background pollers, not from here.
            instances = self.client.instances.all(recursion=1)
            # In a cluster, only show instances running on this node
            if self.local_node:
                instances = [i for i in instances if getattr(i, 'location', None) == self.local_node]
            if not all_tag:
                instances = [i for i in instances if i.status in self.CONTAINER_ACTIVE_STATUS]
            self.display_error = True
        except Exception as e:
            if self.display_error:
                logger.error(f"{self.ext_name} plugin - Can't get instances list ({e})")
                self.display_error = False
            else:
                logger.debug(f"{self.ext_name} plugin - Can't get instances list ({e})")
            return {}, []

        # Start new fetcher threads for new instances
        for instance in instances:
            if instance.name not in self.stats_fetchers:
                logger.debug(f"{self.ext_name} plugin - Create thread for instance {instance.name}")
                self.stats_fetchers[instance.name] = LxdStatsFetcher(instance, poll_interval=self.poll_interval)

        # Stop threads for removed instances
        current_names = {i.name for i in instances}
        absent = set(self.stats_fetchers.keys()) - current_names
        for name in absent:
            logger.debug(f"{self.ext_name} plugin - Stop thread for old instance {name}")
            self.stats_fetchers[name].stop()
            del self.stats_fetchers[name]

        # Generate stats
        container_stats = [self.generate_stats(instance) for instance in instances]
        return {}, container_stats

    @property
    def key(self) -> str:
        return 'name'

    def generate_stats(self, instance) -> dict[str, Any]:
        stats = {
            'key': self.key,
            'name': nativestr(instance.name),
            'id': instance.name,
            'image': instance.config.get('image.description', ''),
            'status': instance.status.lower(),
            'created': instance.created_at,
            'command': None,
            'io': {},
            'cpu': {},
            'memory': {},
            'network': {},
            'io_rx': None,
            'io_wx': None,
            'cpu_percent': None,
            'memory_percent': None,
            'network_rx': None,
            'network_tx': None,
            'ports': '',
            'uptime': None,
        }

        if instance.status not in self.CONTAINER_ACTIVE_STATUS:
            return stats

        if instance.name not in self.stats_fetchers:
            return stats

        stats_fetcher = self.stats_fetchers[instance.name]
        activity_stats = stats_fetcher.activity_stats
        stats.update(activity_stats)

        # Additional fields
        stats['cpu_percent'] = stats['cpu'].get('total')
        stats['cpu_limit'] = stats['cpu'].get('limit')
        stats['memory_usage'] = stats['memory'].get('usage')
        stats['memory_inactive_file'] = stats['memory'].get('inactive_file')
        stats['memory_limit'] = stats['memory'].get('limit')

        if all(k in stats['io'] for k in ('ior', 'iow', 'time_since_update')):
            stats['io_rx'] = stats['io']['ior'] // max(1, stats['io']['time_since_update'])
            stats['io_wx'] = stats['io']['iow'] // max(1, stats['io']['time_since_update'])

        if all(k in stats['network'] for k in ('rx', 'tx', 'time_since_update')):
            stats['network_rx'] = stats['network']['rx'] // max(1, stats['network']['time_since_update'])
            stats['network_tx'] = stats['network']['tx'] // max(1, stats['network']['time_since_update'])

        # Ports from proxy devices (e.g. listen=tcp:0.0.0.0:80 connect=tcp:127.0.0.1:80)
        try:
            devices = instance.expanded_devices or {}
            port_list = []
            for dev in devices.values():
                if dev.get('type') != 'proxy':
                    continue
                listen = dev.get('listen', '')
                connect = dev.get('connect', '')
                # Extract port from "tcp:0.0.0.0:80" format
                listen_port = listen.rsplit(':', 1)[-1] if listen else ''
                connect_port = connect.rsplit(':', 1)[-1] if connect else ''
                if listen_port:
                    proto = listen.split(':')[0] if ':' in listen else 'tcp'
                    port_list.append(f"{listen_port}->{connect_port}/{proto}")
            if port_list:
                stats['ports'] = ','.join(port_list)
        except (AttributeError, TypeError) as e:
            logger.debug(f"{self.ext_name} plugin - Can't get ports for {instance.name} ({e})")

        # Uptime from last_used_at
        try:
            last_used = instance.last_used_at
            if last_used and last_used != '1970-01-01T00:00:00Z':
                started = datetime.fromisoformat(last_used.replace('Z', '+00:00')).replace(tzinfo=None)
                stats['uptime'] = pretty_date(started)
        except (ValueError, AttributeError) as e:
            logger.debug(f"{self.ext_name} plugin - Can't compute uptime for {instance.name} ({e})")

        return stats



#!/usr/bin/env python
#
# Glances - An eye on your system
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Tests for the Load plugin."""

import json

import pytest

from glances.globals import WINDOWS
from glances.plugins.load import load_average, log_core, phys_core


@pytest.fixture
def load_plugin(glances_stats):
    """Return the Load plugin instance from glances_stats."""
    return glances_stats.get_plugin('load')


@pytest.mark.skipif(WINDOWS, reason="Load average not available on Windows")
class TestLoadPluginBasics:
    """Test basic Load plugin functionality."""

    def test_plugin_name(self, load_plugin):
        """Test plugin name is correctly set."""
        assert load_plugin.plugin_name == 'load'

    def test_plugin_is_enabled(self, load_plugin):
        """Test that the plugin is enabled by default."""
        assert load_plugin.is_enabled() is True

    def test_display_curse_enabled(self, load_plugin):
        """Test that curse display is enabled."""
        assert load_plugin.display_curse is True

    def test_history_items_defined(self, load_plugin):
        """Test that history items are properly defined."""
        items = load_plugin.get_items_history_list()
        assert items is not None
        item_names = [item['name'] for item in items]
        assert 'min1' in item_names
        assert 'min5' in item_names
        assert 'min15' in item_names


@pytest.mark.skipif(WINDOWS, reason="Load average not available on Windows")
class TestLoadPluginUpdate:
    """Test Load plugin update functionality."""

    def test_update_returns_dict(self, load_plugin):
        """Test that update returns a dictionary."""
        load_plugin.update()
        stats = load_plugin.get_raw()
        assert isinstance(stats, dict)

    def test_update_contains_mandatory_keys(self, load_plugin):
        """Test that stats contain mandatory keys."""
        load_plugin.update()
        stats = load_plugin.get_raw()
        mandatory_keys = ['min1', 'min5', 'min15', 'cpucore']
        for key in mandatory_keys:
            assert key in stats, f"Missing mandatory key: {key}"

    def test_load_values_non_negative(self, load_plugin):
        """Test that load values are non-negative."""
        load_plugin.update()
        stats = load_plugin.get_raw()
        for key in ['min1', 'min5', 'min15']:
            if key in stats and stats[key] is not None:
                assert stats[key] >= 0, f"{key} should be non-negative"

    def test_cpucore_positive(self, load_plugin):
        """Test that CPU core count is positive."""
        load_plugin.update()
        stats = load_plugin.get_raw()
        assert 'cpucore' in stats
        assert stats['cpucore'] >= 1


@pytest.mark.skipif(WINDOWS, reason="Load average not available on Windows")
class TestLoadPluginViews:
    """Test Load plugin views functionality."""

    def test_update_views_creates_views(self, load_plugin):
        """Test that update_views creates views dictionary."""
        load_plugin.update()
        load_plugin.update_views()
        views = load_plugin.get_views()
        assert isinstance(views, dict)

    def test_views_contain_min15_decoration(self, load_plugin):
        """Test that views contain decoration for min15."""
        load_plugin.update()
        load_plugin.update_views()
        views = load_plugin.get_views()
        if views:  # Views might be empty if no stats
            assert 'min15' in views
            assert 'decoration' in views['min15']

    def test_views_contain_min5_decoration(self, load_plugin):
        """Test that views contain decoration for min5."""
        load_plugin.update()
        load_plugin.update_views()
        views = load_plugin.get_views()
        if views:
            assert 'min5' in views
            assert 'decoration' in views['min5']


@pytest.mark.skipif(WINDOWS, reason="Load average not available on Windows")
class TestLoadPluginJSON:
    """Test Load plugin JSON serialization."""

    def test_get_stats_returns_json(self, load_plugin):
        """Test that get_stats returns valid JSON."""
        load_plugin.update()
        stats_json = load_plugin.get_stats()
        parsed = json.loads(stats_json)
        assert isinstance(parsed, dict)

    def test_json_contains_expected_fields(self, load_plugin):
        """Test that JSON output contains expected fields."""
        load_plugin.update()
        stats_json = load_plugin.get_stats()
        parsed = json.loads(stats_json)
        expected_fields = ['min1', 'min5', 'min15', 'cpucore']
        for field in expected_fields:
            assert field in parsed


@pytest.mark.skipif(WINDOWS, reason="Load average not available on Windows")
class TestLoadPluginHistory:
    """Test Load plugin history functionality."""

    def test_history_enable_check(self, load_plugin):
        """Test that history_enable returns a boolean."""
        result = load_plugin.history_enable()
        assert isinstance(result, bool)

    def test_get_items_history_list(self, load_plugin):
        """Test that get_items_history_list returns the history items."""
        items = load_plugin.get_items_history_list()
        if items is not None:
            assert isinstance(items, list)
            item_names = [item['name'] for item in items]
            assert 'min1' in item_names


@pytest.mark.skipif(WINDOWS, reason="Load average not available on Windows")
class TestLoadPluginReset:
    """Test Load plugin reset functionality."""

    def test_reset_clears_stats(self, load_plugin):
        """Test that reset clears stats."""
        load_plugin.update()
        load_plugin.reset()
        stats = load_plugin.get_raw()
        assert stats == load_plugin.get_init_value()

    def test_reset_views(self, load_plugin):
        """Test that reset_views clears views."""
        load_plugin.update()
        load_plugin.update_views()
        load_plugin.reset_views()
        assert load_plugin.get_views() == {}


@pytest.mark.skipif(WINDOWS, reason="Load average not available on Windows")
class TestLoadPluginFieldsDescription:
    """Test Load plugin fields description."""

    def test_fields_description_exists(self, load_plugin):
        """Test that fields_description is defined."""
        assert load_plugin.fields_description is not None

    def test_mandatory_fields_described(self, load_plugin):
        """Test that mandatory fields have descriptions."""
        mandatory_fields = ['min1', 'min5', 'min15', 'cpucore']
        for field in mandatory_fields:
            assert field in load_plugin.fields_description

    def test_field_has_description(self, load_plugin):
        """Test that each field has a description."""
        for field, info in load_plugin.fields_description.items():
            assert 'description' in info, f"Field {field} missing description"


@pytest.mark.skipif(WINDOWS, reason="Load average not available on Windows")
class TestLoadPluginMsgCurse:
    """Test Load plugin curse message generation."""

    def test_msg_curse_returns_list(self, load_plugin):
        """Test that msg_curse returns a list."""
        load_plugin.update()
        msg = load_plugin.msg_curse()
        assert isinstance(msg, list)

    def test_msg_curse_format(self, load_plugin):
        """Test that msg_curse returns properly formatted entries."""
        load_plugin.update()
        msg = load_plugin.msg_curse()
        if msg:
            for entry in msg:
                assert isinstance(entry, dict)
                assert 'msg' in entry

    def test_msg_curse_structure(self, load_plugin):
        """Test msg_curse output structure when stats available."""
        load_plugin.update()
        stats = load_plugin.get_raw()
        if stats and not load_plugin.is_disabled():
            msg = load_plugin.msg_curse()
            if msg:
                assert len(msg) > 0


class TestLoadHelperFunctions:
    """Test Load plugin helper functions."""

    def test_log_core_returns_int(self):
        """Test that log_core returns an integer."""
        cores = log_core()
        assert isinstance(cores, int)
        assert cores >= 1

    def test_phys_core_returns_int(self):
        """Test that phys_core returns an integer."""
        cores = phys_core()
        assert isinstance(cores, int)
        assert cores >= 1

    @pytest.mark.skipif(WINDOWS, reason="Load average not available on Windows")
    def test_load_average_returns_tuple(self):
        """Test that load_average returns a tuple."""
        load = load_average()
        if load is not None:
            assert isinstance(load, tuple)
            assert len(load) == 3

    @pytest.mark.skipif(WINDOWS, reason="Load average not available on Windows")
    def test_load_average_values_non_negative(self):
        """Test that load_average values are non-negative."""
        load = load_average()
        if load is not None:
            for value in load:
                assert value >= 0

    @pytest.mark.skipif(WINDOWS, reason="Load average not available on Windows")
    def test_load_average_percent_mode(self):
        """Test load_average in percent mode."""
        load = load_average(percent=True)
        if load is not None:
            assert isinstance(load, tuple)
            assert len(load) == 3
            # Values can exceed 100% if load is high


@pytest.mark.skipif(WINDOWS, reason="Load average not available on Windows")
class TestLoadPluginAlerts:
    """Test Load plugin alert functionality."""

    def test_alert_based_on_cpucore(self, load_plugin):
        """Test that alerts are calculated based on CPU cores."""
        load_plugin.update()
        stats = load_plugin.get_raw()
        if 'cpucore' in stats and 'min15' in stats:
            # Maximum is 100 * cpucore
            alert = load_plugin.get_alert_log(stats['min15'], maximum=100 * stats['cpucore'])
            assert alert is not None


@pytest.mark.skipif(WINDOWS, reason="Load average not available on Windows")
class TestLoadPluginExport:
    """Test Load plugin export functionality."""

    def test_get_export_returns_dict(self, load_plugin):
        """Test that get_export returns a dict."""
        load_plugin.update()
        export = load_plugin.get_export()
        assert isinstance(export, dict)

    def test_export_equals_raw(self, load_plugin):
        """Test that export equals raw stats by default."""
        load_plugin.update()
        assert load_plugin.get_export() == load_plugin.get_raw()

    def test_export_contains_load_values_when_available(self, load_plugin):
        """Test that export contains load values when available."""
        load_plugin.update()
        export = load_plugin.get_export()
        if export:
            assert 'min1' in export
            assert 'min5' in export
            assert 'min15' in export



#
# This file is part of Glances.
#
# SPDX-FileCopyrightText: 2024 Nicolas Hennion <nicolas@nicolargo.com>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Generate Glances Restful API documentation."""

import json
import time
from pprint import pformat

from glances import __apiversion__
from glances.logger import logger

API_URL = f"http://localhost:61208/api/{__apiversion__}"

APIDOC_HEADER = f"""\
.. _api_restful:

Restful/JSON API documentation
==============================

This documentation describes the Glances API version {__apiversion__} (Restful/JSON) interface.

An OpenAPI specification file is available at:
``https://raw.githubusercontent.com/nicolargo/glances/refs/heads/develop/docs/api/openapi.json``

Run the Glances API server
--------------------------

The Glances Restful/API server could be ran using the following command line:

.. code-block:: bash

    # glances -w --disable-webui

It is also ran automatically when Glances is started in Web server mode (-w).

If you want to enable the Glances Central Browser, use:

.. code-block:: bash

    # glances -w --browser --disable-webui

API URL
-------

The default root API URL is ``http://localhost:61208/api/{__apiversion__}``.

The bind address and port could be changed using the ``--bind`` and ``--port`` command line options.

It is also possible to define an URL prefix using the ``url_prefix`` option from the [outputs] section
of the Glances configuration file.

Note: The url_prefix should always end with a slash (``/``).

For example:

.. code-block:: ini

    [outputs]
    url_prefix = /glances/

will change the root API URL to ``http://localhost:61208/glances/api/{__apiversion__}`` and the Web UI URL to
``http://localhost:61208/glances/``

API documentation URL
---------------------

The API documentation is embeded in the server and available at the following URL:
``http://localhost:61208/docs#/``.

Authentication
--------------

Glances API supports both HTTP Basic authentication and JWT (JSON Web Token) Bearer authentication.

To enable authentication, start Glances with the ``--password`` option.

To generate a new login/password pair, use the following command:

.. code-block:: bash

    glances -w --username
    > Enter new username: foo
    > Enter new password: ********
    > Confirm new password: ********
    > User 'username' created/updated successfully.

To reuse an existing login/password pair, start Glances with the ``-u <user>`` option:

.. code-block:: bash

    glances -w -u foo

JWT Token Authentication
~~~~~~~~~~~~~~~~~~~~~~~~

JWT authentication requires the ``python-jose`` library to be installed.

**Step 1: Get a JWT Token**

Request a token by sending your credentials to the token endpoint:

.. code-block:: bash

    curl -X POST http://localhost:61208/api/{__apiversion__}/token \\
      -H "Content-Type: application/json" \\
      -d '{{"username": "your_username", "password": "your_password"}}'

This will return a response like:

.. code-block:: json

    {{
      "access_token": "...",
      "token_type": "bearer",
      "expires_in": 3600
    }}

**Step 2: Use the Token**

Use the token in the Authorization header with Bearer authentication:

.. code-block:: bash

    # Store the token in a variable
    TOKEN="your_access_token_here"

    # Access a protected endpoint
    curl -H "Authorization: Bearer $TOKEN" \\
      http://localhost:61208/api/{__apiversion__}/cpu

**Complete Example:**

.. code-block:: bash

    # Get token and extract access_token
    TOKEN=$(curl -s -X POST http://localhost:61208/api/{__apiversion__}/token \\
      -H "Content-Type: application/json" \\
      -d '{{"username": "glances", "password": "mypassword"}}' \\
      | grep -o '"access_token":"[^"]*"' \\
      | cut -d'"' -f4)

    # Use the token to get CPU stats
    curl -H "Authorization: Bearer $TOKEN" \\
      http://localhost:61208/api/{__apiversion__}/cpu

**Configuration:**

You can configure JWT settings in the Glances configuration file:

.. code-block:: ini

    [outputs]
    # JWT secret key (if not set, a random key will be generated)
    jwt_secret_key = your-secret-key-here
    # JWT token expiration in minutes (default: 60)
    jwt_expire_minutes = 60

**Note:** The token endpoint (``/api/{__apiversion__}/token``) does not require authentication.
Protected endpoints support both Bearer token and Basic Auth authentication methods.

.. _security:

Security
--------

By default, Glances web server runs **without authentication** and binds to
**all network interfaces** (``0.0.0.0``). This means any client that can reach
the server on the network can access the full REST API, including sensitive
system information such as process command-lines, which may contain credentials
(passwords, API keys, tokens passed as arguments).

This default is intentional for ease of use on private, trusted networks (home
labs, local machines, internal infrastructure). However, if your Glances
instance is reachable from untrusted networks, you should take the following
precautions:

**Enable authentication** by starting Glances with the ``--password`` option:

.. code-block:: bash

    glances -w --password

**Bind to localhost only** if remote access is not needed:

.. code-block:: bash

    glances -w --bind 127.0.0.1

**Enable DNS rebinding protection** by setting ``webui_allowed_hosts`` in
``glances.conf``. This restricts the HTTP ``Host`` header values accepted by the
web server. Without this setting, a DNS rebinding attack could allow an
untrusted web page to read the REST API from a victim's browser on the same
network, even without direct network access to the Glances instance.

.. code-block:: ini

    [outputs]
    # Comma-separated list of allowed hostnames/IPs.
    # Wildcards are supported (e.g. *.example.com).
    webui_allowed_hosts=localhost,127.0.0.1,myserver.example.com

When ``webui_allowed_hosts`` is set, requests with a ``Host`` header not in the
list are rejected with ``400 Bad Request``. When absent or commented out
(the default), no host filtering is applied.

**Use a reverse proxy** (nginx, Caddy, Apache) with TLS and authentication for
any public-facing or semi-public deployment. This is the recommended approach
for production environments.

.. code-block:: ini

    # Example: restrict bind to localhost, access via reverse proxy
    # In glances.conf:
    [outputs]
    # Bind to localhost, let the reverse proxy handle external access
    # then configure your reverse proxy to forward to 127.0.0.1:61208
    webui_allowed_hosts=localhost,127.0.0.1

.. note::

    The bind address (``0.0.0.0`` by default) controls which network interfaces
    the server listens on, but it is **not a security boundary**. For
    deployments on non-loopback interfaces, always set ``webui_allowed_hosts``
    and consider enabling authentication.

**CORS (Cross-Origin Resource Sharing)** controls which external websites can
make requests to the Glances API from a browser. By default, Glances allows
requests from any origin (``cors_origins=*``) but does **not** allow credentials
(``cors_credentials=False``). This means cross-origin requests work for
unauthenticated API access, but browsers will not send stored credentials
(e.g. Basic Auth) to the API from a third-party page.

If you need credentialed cross-origin access (e.g. a separate dashboard
application that authenticates to Glances), you **must** configure explicit
origins — the wildcard ``*`` combined with credentials is insecure and will be
automatically rejected:

.. code-block:: ini

    [outputs]
    cors_origins=https://my-dashboard.internal.example.com
    cors_credentials=True

.. warning::

    Setting ``cors_credentials=True`` with ``cors_origins=*`` is not allowed.
    Glances will automatically disable credentials and log a warning if this
    combination is detected. This prevents a class of cross-site data theft
    attacks where any website could read your monitoring data.

The same ``cors_origins`` key also applies to the legacy XML-RPC server
(``glances -s``): when multiple origins are listed, each request's ``Origin``
header is compared against the allowlist and reflected only if it is an exact
match (``Vary: Origin`` is set). The default ``*`` keeps the wildcard
behaviour for backward compatibility.

When Glances is started without authentication or without host filtering,
warning messages are displayed at startup to remind you of the risks.

WebUI refresh
-------------

It is possible to change the Web UI refresh rate (default is 2 seconds) using the following option in the URL:
``http://localhost:61208/?refresh=5``

"""


def indent_stat(stat, indent='    '):
    # Indent stats to pretty print it
    if isinstance(stat, list) and len(stat) > 1 and isinstance(stat[0], dict):
        # Only display two first items
        return indent + pformat(stat[0:2]).replace('\n', '\n' + indent).replace("'", '"')
    return indent + pformat(stat).replace('\n', '\n' + indent).replace("'", '"')


def print_api_status():
    sub_title = 'GET API status'
    print(sub_title)
    print('-' * len(sub_title))
    print('')
    print('This entry point should be used to check the API status.')
    print('It will the Glances version and a 200 return code if everything is OK.')
    print('')
    print('Get the Rest API status::')
    print('')
    print(f'    # curl -I {API_URL}/status')
    print(indent_stat('HTTP/1.0 200 OK'))
    print('')


def print_plugins_list(stat):
    sub_title = 'GET plugins list'
    print(sub_title)
    print('-' * len(sub_title))
    print('')
    print('Get the plugins list::')
    print('')
    print(f'    # curl {API_URL}/pluginslist')
    print(indent_stat(stat))
    print('')


def print_plugin_stats(plugin, stat):
    sub_title = f'GET {plugin}'
    print(sub_title)
    print('-' * len(sub_title))
    print('')

    print('Get plugin stats::')
    print('')
    print(f'    # curl {API_URL}/{plugin}')
    print(indent_stat(json.loads(stat.get_stats())))
    print('')


def print_plugin_description(plugin, stat):
    if stat.fields_description:
        # For each plugins with a description
        print('Fields descriptions:')
        print('')
        time_since_update = False
        for field, description in stat.fields_description.items():
            print(
                '* **{}**: {} (unit is *{}*)'.format(
                    field,
                    (
                        description['description'][:-1]
                        if description['description'].endswith('.')
                        else description['description']
                    ),
                    description['unit'] if 'unit' in description else 'None',
                )
            )
            if 'rate' in description and description['rate']:
                time_since_update = True
                print(
                    '* **{}**: {} (unit is *{}* per second)'.format(
                        field + '_rate_per_sec',
                        (
                            description['description'][:-1]
                            if description['description'].endswith('.')
                            else description['description']
                        )
                        + ' per second',
                        description['unit'] if 'unit' in description else 'None',
                    )
                )
                print(
                    '* **{}**: {} (unit is *{}*)'.format(
                        field + '_gauge',
                        (
                            description['description'][:-1]
                            if description['description'].endswith('.')
                            else description['description']
                        )
                        + ' (cumulative)',
                        description['unit'] if 'unit' in description else 'None',
                    )
                )

        if time_since_update:
            print(
                '* **{}**: {} (unit is *{}*)'.format(
                    'time_since_update', 'Number of seconds since last update', 'seconds'
                )
            )

        print('')
    else:
        logger.error(f'No fields_description variable defined for plugin {plugin}')


def print_plugin_item_value(plugin, stat, stat_export):
    item = None
    value = None
    if isinstance(stat_export, dict):
        item = list(stat_export.keys())[0]
        value = None
    elif isinstance(stat_export, list) and len(stat_export) > 0 and isinstance(stat_export[0], dict):
        if 'key' in stat_export[0]:
            item = stat_export[0]['key']
        else:
            item = list(stat_export[0].keys())[0]
    if item and stat.get_stats_item(item):
        stat_item = json.loads(stat.get_stats_item(item))
        if isinstance(stat_item[item], list):
            value = stat_item[item][0]
        else:
            value = stat_item[item]
        print('Get a specific field::')
        print('')
        print(f'    # curl {API_URL}/{plugin}/{item}')
        print(indent_stat(stat_item))
        print('')
    if item and value and stat.get_stats_value(item, value):
        print('Get a specific item when field matches the given value::')
        print('')
        print(f'    # curl {API_URL}/{plugin}/{item}/value/{value}')
        print(indent_stat(json.loads(stat.get_stats_value(item, value))))
        print('')


def print_all():
    sub_title = 'GET all stats'
    print(sub_title)
    print('-' * len(sub_title))
    print('')
    print('Get all Glances stats::')
    print('')
    print(f'    # curl {API_URL}/all')
    print('    Return a very big dictionary with all stats')
    print('')
    print('Note: Update is done automatically every time /all or /<plugin> is called.')
    print('')


def print_processes():
    sub_title = 'GET stats of a specific process'
    print(sub_title)
    print('-' * len(sub_title))
    print('')
    print('Get stats for process with PID == 777::')
    print('')
    print(f'    # curl {API_URL}/processes/777')
    print('    Return stats for process (dict)')
    print('')
    print('Enable extended stats for process with PID == 777 (only one process at a time can be enabled)::')
    print('')
    print(f'    # curl -X POST {API_URL}/processes/extended/777')
    print(f'    # curl {API_URL}/all')
    print(f'    # curl {API_URL}/processes/777')
    print('    Return stats for process (dict)')
    print('')
    print('Note: Update *is not* done automatically when you call /processes/<pid>.')
    print('')


def print_top(stats):
    time.sleep(1)
    stats.update()
    sub_title = 'GET top n items of a specific plugin'
    print(sub_title)
    print('-' * len(sub_title))
    print('')
    print('Get top 2 processes of the processlist plugin::')
    print('')
    print(f'    # curl {API_URL}/processlist/top/2')
    print(indent_stat(stats.get_plugin('processlist').get_export()[:2]))
    print('')
    print('Note: Only work for plugin with a list of items')
    print('')


def print_fields_info(stats):
    sub_title = 'GET item description'
    print(sub_title)
    print('-' * len(sub_title))
    print('Get item description (human readable) for a specific plugin/item::')
    print('')
    print(f'    # curl {API_URL}/diskio/read_bytes/description')
    print(indent_stat(stats.get_plugin('diskio').get_item_info('read_bytes', 'description')))
    print('')
    print('Note: the description is defined in the fields_description variable of the plugin.')
    print('')
    sub_title = 'GET item unit'
    print(sub_title)
    print('-' * len(sub_title))
    print('Get item unit for a specific plugin/item::')
    print('')
    print(f'    # curl {API_URL}/diskio/read_bytes/unit')
    print(indent_stat(stats.get_plugin('diskio').get_item_info('read_bytes', 'unit')))
    print('')
    print('Note: the description is defined in the fields_description variable of the plugin.')
    print('')


def print_history(stats):
    time.sleep(1)
    stats.update()
    time.sleep(1)
    stats.update()
    sub_title = 'GET stats history'
    print(sub_title)
    print('-' * len(sub_title))
    print('')
    print('History of a plugin::')
    print('')
    print(f'    # curl {API_URL}/cpu/history')
    print(indent_stat(json.loads(stats.get_plugin('cpu').get_stats_history(nb=3))))
    print('')
    print('Limit history to last 2 values::')
    print('')
    print(f'    # curl {API_URL}/cpu/history/2')
    print(indent_stat(json.loads(stats.get_plugin('cpu').get_stats_history(nb=2))))
    print('')
    print('History for a specific field::')
    print('')
    print(f'    # curl {API_URL}/cpu/system/history')
    print(indent_stat(json.loads(stats.get_plugin('cpu').get_stats_history('system'))))
    print('')
    print('Limit history for a specific field to last 2 values::')
    print('')
    print(f'    # curl {API_URL}/cpu/system/history')
    print(indent_stat(json.loads(stats.get_plugin('cpu').get_stats_history('system', nb=2))))
    print('')


def print_limits(stats):
    sub_title = 'GET limits (used for thresholds)'
    print(sub_title)
    print('-' * len(sub_title))
    print('')
    print('All limits/thresholds::')
    print('')
    print(f'    # curl {API_URL}/all/limits')
    print(indent_stat(stats.getAllLimitsAsDict()))
    print('')
    print('Limits/thresholds for the cpu plugin::')
    print('')
    print(f'    # curl {API_URL}/cpu/limits')
    print(indent_stat(stats.get_plugin('cpu').limits))
    print('')


def print_plugin_post_events():
    sub_title = 'POST clear events'
    print(sub_title)
    print('-' * len(sub_title))
    print('')
    print('Clear all alarms from the list::')
    print('')
    print(f'    # curl -H "Content-Type: application/json" -X POST {API_URL}/events/clear/all')
    print('')
    print('Clear warning alarms from the list::')
    print('')
    print(f'    # curl -H "Content-Type: application/json" -X POST {API_URL}/events/clear/warning')
    print('')


class GlancesStdoutApiRestfulDoc:
    """This class manages the fields description display."""

    def __init__(self, config=None, args=None):
        # Init
        self.config = config
        self.args = args

    def end(self):
        pass

    def update(self, stats, duration=1):
        """Display issue"""

        # Display header
        print(APIDOC_HEADER)

        # Display API status
        print_api_status()

        # Display plugins list
        print_plugins_list(sorted(stats._plugins))

        # Loop over plugins
        for plugin in sorted(stats._plugins):
            stat = stats.get_plugin(plugin)
            print_plugin_stats(plugin, stat)
            print_plugin_description(plugin, stat)
            if plugin == 'alert':
                print_plugin_post_events()

            stat_export = stat.get_export()
            if stat_export is None or stat_export == [] or stat_export == {}:
                continue
            print_plugin_item_value(plugin, stat, stat_export)

        # Get all stats
        print_all()

        # Get process stats
        print_processes()

        # Get top stats (only for plugins with a list of items)
        # Example for processlist plugin: get top 2 processes
        print_top(stats)

        # Fields description
        print_fields_info(stats)

        # History
        print_history(stats)

        # Limits
        print_limits(stats)

        # Return True to exit directly (no refresh)
        return True



#!/usr/bin/env python
#
# Glances - An eye on your system
#
# SPDX-FileCopyrightText: 2026 Christian Rishøj <christian@rishoj.net>
#
# SPDX-License-Identifier: LGPL-3.0-only
#

"""Tests for the LXD container engine."""

import time
from types import SimpleNamespace
from unittest.mock import MagicMock, patch

import pytest


def make_mock_state(
    cpu_ns=5_000_000_000,
    mem_usage=1_000_000_000,
    mem_total=2_000_000_000,
    mem_usage_peak=None,
    disk_usage=500_000,
    net_rx=10_000,
    net_tx=20_000,
):
    """Create a mock LXD instance state."""
    return SimpleNamespace(
        cpu={"usage": cpu_ns},
        memory={"usage": mem_usage, "total": mem_total, "usage_peak": mem_usage_peak or mem_total},
        disk={"root": {"usage": disk_usage}},
        network={
            "eth0": {
                "type": "broadcast",
                "counters": {"bytes_received": net_rx, "bytes_sent": net_tx},
            },
            "lo": {
                "type": "loopback",
                "counters": {"bytes_received": 999, "bytes_sent": 999},
            },
        },
    )


def make_mock_instance(
    name="test-container",
    status="Running",
    config=None,
    expanded_config=None,
    expanded_devices=None,
    created_at="2026-01-01T00:00:00Z",
    last_used_at="2026-03-15T10:00:00Z",
    location="node1",
):
    """Create a mock LXD instance."""
    instance = MagicMock()
    instance.name = name
    instance.status = status
    instance.config = config or {"image.description": "Ubuntu 24.04 LTS"}
    instance.expanded_config = expanded_config or {}
    instance.expanded_devices = expanded_devices or {}
    instance.created_at = created_at
    instance.last_used_at = last_used_at
    instance.location = location
    return instance


@pytest.fixture
def mock_instance():
    """Return a mock LXD instance that returns incrementing CPU stats."""
    instance = make_mock_instance()
    call_count = 0

    def stateful_state():
        nonlocal call_count
        call_count += 1
        return make_mock_state(
            cpu_ns=5_000_000_000 * call_count,
            net_rx=10_000 * call_count,
            net_tx=20_000 * call_count,
            disk_usage=500_000 * call_count,
        )

    instance.state = stateful_state
    return instance


class TestLxdStatsFetcher:
    """Test LxdStatsFetcher stat computation."""

    def test_no_state_returns_empty(self, mock_instance):
        from glances.plugins.containers.engines.lxd import LxdStatsFetcher

        fetcher = LxdStatsFetcher(mock_instance, poll_interval=999)
        # Before any poll completes, force state to None
        fetcher._state = None
        stats = fetcher.activity_stats
        fetcher.stop()

        assert stats == {"cpu": {}, "memory": {}, "io": {}, "network": {}}

    def test_first_poll_returns_zero_cpu(self, mock_instance):
        from glances.plugins.containers.engines.lxd import LxdStatsFetcher

        fetcher = LxdStatsFetcher(mock_instance, poll_interval=999)
        # Wait for first poll
        time.sleep(0.2)
        stats = fetcher.activity_stats
        fetcher.stop()

        assert stats["cpu"]["total"] == 0.0
        assert stats["memory"]["usage"] == 1_000_000_000
        assert stats["memory"]["limit"] == 2_000_000_000

    def test_second_poll_computes_cpu_delta(self, mock_instance):
        from glances.plugins.containers.engines.lxd import LxdStatsFetcher

        fetcher = LxdStatsFetcher(mock_instance, poll_interval=999)
        time.sleep(0.2)

        # First reading (baseline)
        fetcher.activity_stats

        # Simulate second poll
        fetcher._state = mock_instance.state()
        stats = fetcher.activity_stats
        fetcher.stop()

        assert "total" in stats["cpu"]
        assert stats["cpu"]["total"] > 0

    def test_network_excludes_loopback(self, mock_instance):
        from glances.plugins.containers.engines.lxd import LxdStatsFetcher

        fetcher = LxdStatsFetcher(mock_instance, poll_interval=999)
        time.sleep(0.2)
        stats = fetcher.activity_stats
        fetcher.stop()

        # Should only count eth0, not lo
        assert stats["network"]["cumulative_rx"] == 10_000
        assert stats["network"]["cumulative_tx"] == 20_000

    def test_memory_unlimited_falls_back_to_peak(self):
        from glances.plugins.containers.engines.lxd import LxdStatsFetcher

        instance = make_mock_instance()
        instance.state = lambda: make_mock_state(mem_total=0, mem_usage_peak=4_000_000_000)

        fetcher = LxdStatsFetcher(instance, poll_interval=999)
        time.sleep(0.2)
        stats = fetcher.activity_stats
        fetcher.stop()

        # total=0 means unlimited, should fall back to usage_peak
        assert stats["memory"]["limit"] == 4_000_000_000

    def test_cpu_limit_explicit(self):
        from glances.plugins.containers.engines.lxd import LxdStatsFetcher

        instance = make_mock_instance(expanded_config={'limits.cpu': '4'})
        instance.state = lambda: make_mock_state()

        fetcher = LxdStatsFetcher(instance, poll_interval=999)
        time.sleep(0.2)
        stats = fetcher.activity_stats
        fetcher.stop()

        assert stats["cpu"]["limit"] == 4.0

    def test_cpu_limit_fallback(self):
        from glances.plugins.containers.engines.lxd import LxdStatsFetcher

        instance = make_mock_instance(expanded_config={})
        instance.state = lambda: make_mock_state()

        with patch('psutil.cpu_count', return_value=8):
            fetcher = LxdStatsFetcher(instance, poll_interval=999)
            time.sleep(0.2)
            stats = fetcher.activity_stats
            fetcher.stop()

        assert stats["cpu"]["limit"] == 8.0

    def test_stop_terminates_thread(self, mock_instance):
        from glances.plugins.containers.engines.lxd import LxdStatsFetcher

        fetcher = LxdStatsFetcher(mock_instance, poll_interval=999)
        fetcher.stop()
        time.sleep(0.1)
        assert not fetcher._thread.is_alive()


class TestLxdExtensionGenerateStats:
    """Test LxdExtension.generate_stats with mock instances."""

    def _make_extension(self):
        """Create an LxdExtension without connecting to a real server."""
        from glances.plugins.containers.engines.lxd import LxdExtension

        with patch.object(LxdExtension, '__init__', lambda self, **kwargs: None):
            ext = LxdExtension.__new__(LxdExtension)
            ext.ext_name = "containers (LXD)"
            ext.stats_fetchers = {}
            ext.CONTAINER_ACTIVE_STATUS = ['Running']
            return ext

    def test_stopped_instance_returns_minimal_stats(self):
        ext = self._make_extension()
        instance = make_mock_instance(status="Stopped")
        stats = ext.generate_stats(instance)

        assert stats['name'] == 'test-container'
        assert stats['status'] == 'stopped'
        assert stats['cpu_percent'] is None
        assert 'memory_usage' not in stats

    def test_running_instance_with_fetcher(self):
        ext = self._make_extension()
        instance = make_mock_instance()

        fetcher = MagicMock()
        fetcher.activity_stats = {
            "cpu": {"total": 50.0, "limit": 2.0},
            "memory": {"usage": 1_000_000_000, "limit": 2_000_000_000, "inactive_file": 0},
            "io": {"ior": 100, "iow": 50, "time_since_update": 2},
            "network": {"rx": 5000, "tx": 3000, "time_since_update": 2},
        }
        ext.stats_fetchers["test-container"] = fetcher

        stats = ext.generate_stats(instance)

        assert stats['cpu_percent'] == 50.0
        assert stats['cpu_limit'] == 2.0
        assert stats['memory_usage'] == 1_000_000_000
        assert stats['memory_limit'] == 2_000_000_000
        assert stats['io_rx'] == 50
        assert stats['io_wx'] == 25
        assert stats['network_rx'] == 2500
        assert stats['network_tx'] == 1500

    def test_proxy_device_ports(self):
        ext = self._make_extension()
        instance = make_mock_instance(
            expanded_devices={
                "http": {"type": "proxy", "listen": "tcp:0.0.0.0:80", "connect": "tcp:127.0.0.1:8080"},
                "https": {"type": "proxy", "listen": "tcp:0.0.0.0:443", "connect": "tcp:127.0.0.1:8443"},
                "root": {"type": "disk", "path": "/", "pool": "default"},
            },
        )

        fetcher = MagicMock()
        fetcher.activity_stats = {"cpu": {}, "memory": {}, "io": {}, "network": {}}
        ext.stats_fetchers["test-container"] = fetcher

        stats = ext.generate_stats(instance)

        assert "80->8080/tcp" in stats['ports']
        assert "443->8443/tcp" in stats['ports']

    def test_no_proxy_devices_empty_ports(self):
        ext = self._make_extension()
        instance = make_mock_instance(
            expanded_devices={"root": {"type": "disk", "path": "/", "pool": "default"}},
        )

        fetcher = MagicMock()
        fetcher.activity_stats = {"cpu": {}, "memory": {}, "io": {}, "network": {}}
        ext.stats_fetchers["test-container"] = fetcher

        stats = ext.generate_stats(instance)
        assert stats['ports'] == ''

    def test_image_from_config(self):
        ext = self._make_extension()
        instance = make_mock_instance(config={"image.description": "Alpine 3.19"})
        stats = ext.generate_stats(instance)
        assert stats['image'] == 'Alpine 3.19'


class TestLxdExtensionUpdate:
    """Test LxdExtension.update with mock client."""

    def _make_extension_with_client(self, instances, local_node=None):
        from glances.plugins.containers.engines.lxd import LxdExtension

        with patch.object(LxdExtension, '__init__', lambda self, **kwargs: None):
            ext = LxdExtension.__new__(LxdExtension)
            ext.ext_name = "containers (LXD)"
            ext.stats_fetchers = {}
            ext.CONTAINER_ACTIVE_STATUS = ['Running']
            ext.display_error = True
            ext.disable = False
            ext.poll_interval = 999
            ext.local_node = local_node
            ext.client = MagicMock()
            ext.client.instances.all.return_value = instances
            return ext

    def test_filters_to_running_only(self):
        running = make_mock_instance(name="web", status="Running")
        stopped = make_mock_instance(name="db", status="Stopped")
        ext = self._make_extension_with_client([running, stopped])

        with patch('glances.plugins.containers.engines.lxd.LxdStatsFetcher'):
            _, container_stats = ext.update(all_tag=False)

        assert len(container_stats) == 1
        assert container_stats[0]['name'] == 'web'

    def test_all_tag_includes_stopped(self):
        running = make_mock_instance(name="web", status="Running")
        stopped = make_mock_instance(name="db", status="Stopped")
        ext = self._make_extension_with_client([running, stopped])

        with patch('glances.plugins.containers.engines.lxd.LxdStatsFetcher'):
            _, container_stats = ext.update(all_tag=True)

        assert len(container_stats) == 2

    def test_cluster_filters_to_local_node(self):
        local = make_mock_instance(name="web", location="node1")
        remote = make_mock_instance(name="db", location="node2")
        ext = self._make_extension_with_client([local, remote], local_node="node1")

        with patch('glances.plugins.containers.engines.lxd.LxdStatsFetcher'):
            _, container_stats = ext.update(all_tag=True)

        assert len(container_stats) == 1
        assert container_stats[0]['name'] == 'web'

    def test_standalone_does_not_filter_by_location(self):
        # On a non-clustered LXD host, instance.location is typically empty.
        # local_node stays None, so every instance must pass through.
        a = make_mock_instance(name="web", location="")
        b = make_mock_instance(name="db", location=None)
        ext = self._make_extension_with_client([a, b], local_node=None)

        with patch('glances.plugins.containers.engines.lxd.LxdStatsFetcher'):
            _, container_stats = ext.update(all_tag=True)

        assert {c['name'] for c in container_stats} == {"web", "db"}  # nosec B101

    def test_cleans_up_removed_instances(self):
        instance = make_mock_instance(name="web")
        ext = self._make_extension_with_client([instance])

        mock_fetcher = MagicMock()
        ext.stats_fetchers["old-container"] = mock_fetcher

        with patch('glances.plugins.containers.engines.lxd.LxdStatsFetcher'):
            ext.update(all_tag=True)

        mock_fetcher.stop.assert_called_once()
        assert "old-container" not in ext.stats_fetchers

    def test_disabled_returns_empty(self):
        ext = self._make_extension_with_client([])
        ext.disable = True
        version, containers = ext.update(all_tag=True)
        assert version == {}
        assert containers == []


class TestLxdExtensionConnect:  # noqa: D203
    """Test LxdExtension.connect detection of cluster membership."""

    def _make_extension_for_connect(self, host_info):
        from glances.plugins.containers.engines.lxd import LxdExtension

        mock_client = MagicMock()
        mock_client.host_info = host_info

        with patch.object(LxdExtension, '__init__', lambda self, **kwargs: None):
            ext = LxdExtension.__new__(LxdExtension)
            ext.ext_name = "containers (LXD)"
            ext.endpoint = None
            ext.local_node = None
            ext.disable = False

            with patch('glances.plugins.containers.engines.lxd.LxdClient', return_value=mock_client, create=True):
                ext.connect()
        return ext

    def test_standalone_leaves_local_node_unset(self):
        ext = self._make_extension_for_connect({"environment": {"server_name": "zfs01", "server_clustered": False}})
        assert ext.local_node is None  # nosec B101

    def test_cluster_sets_local_node(self):
        ext = self._make_extension_for_connect({"environment": {"server_name": "node1", "server_clustered": True}})
        assert ext.local_node == "node1"  # nosec B101
